import logging

from .default_permissions_endpoint import _DefaultPermissionsEndpoint
from .endpoint import QuerysetEndpoint, api, XML_CONTENT_TYPE
from .exceptions import MissingRequiredFieldError
from .permissions_endpoint import _PermissionsEndpoint
from tableauserverclient.server import RequestFactory, RequestOptions
from tableauserverclient.models import ProjectItem, PaginationItem, Resource

from typing import List, Optional, Tuple, TYPE_CHECKING

if TYPE_CHECKING:
    from ..server import Server
    from ..request_options import RequestOptions

logger = logging.getLogger("tableau.endpoint.projects")


class Projects(QuerysetEndpoint):
    def __init__(self, parent_srv: "Server") -> None:
        super(Projects, self).__init__(parent_srv)

        self._permissions = _PermissionsEndpoint(parent_srv, lambda: self.baseurl)
        self._default_permissions = _DefaultPermissionsEndpoint(parent_srv, lambda: self.baseurl)

    @property
    def baseurl(self) -> str:
        return "{0}/sites/{1}/projects".format(self.parent_srv.baseurl, self.parent_srv.site_id)

    @api(version="2.0")
    def get(self, req_options: Optional["RequestOptions"] = None) -> Tuple[List[ProjectItem], PaginationItem]:
        logger.info("Querying all projects on site")
        url = self.baseurl
        server_response = self.get_request(url, req_options)
        pagination_item = PaginationItem.from_response(server_response.content, self.parent_srv.namespace)
        all_project_items = ProjectItem.from_response(server_response.content, self.parent_srv.namespace)
        return all_project_items, pagination_item

    @api(version="2.0")
    def delete(self, project_id: str) -> None:
        if not project_id:
            error = "Project ID undefined."
            raise ValueError(error)
        url = "{0}/{1}".format(self.baseurl, project_id)
        self.delete_request(url)
        logger.info("Deleted single project (ID: {0})".format(project_id))

    @api(version="2.0")
    def update(self, project_item: ProjectItem, samples: bool = False) -> ProjectItem:
        if not project_item.id:
            error = "Project item missing ID."
            raise MissingRequiredFieldError(error)

        params = {"params": {RequestOptions.Field.PublishSamples: samples}}
        url = "{0}/{1}".format(self.baseurl, project_item.id)
        update_req = RequestFactory.Project.update_req(project_item)
        server_response = self.put_request(url, update_req, XML_CONTENT_TYPE, params)
        logger.info("Updated project item (ID: {0})".format(project_item.id))
        updated_project = ProjectItem.from_response(server_response.content, self.parent_srv.namespace)[0]
        return updated_project

    @api(version="2.0")
    def create(self, project_item: ProjectItem, samples: bool = False) -> ProjectItem:
        params = {"params": {RequestOptions.Field.PublishSamples: samples}}
        url = self.baseurl
        create_req = RequestFactory.Project.create_req(project_item)
        server_response = self.post_request(url, create_req, XML_CONTENT_TYPE, params)
        new_project = ProjectItem.from_response(server_response.content, self.parent_srv.namespace)[0]
        logger.info("Created new project (ID: {0})".format(new_project.id))
        return new_project

    @api(version="2.0")
    def populate_permissions(self, item: ProjectItem) -> None:
        self._permissions.populate(item)

    @api(version="2.0")
    def update_permission(self, item, rules):
        import warnings

        warnings.warn(
            "Server.projects.update_permission is deprecated, "
            "please use Server.projects.update_permissions instead.",
            DeprecationWarning,
        )
        return self._permissions.update(item, rules)

    @api(version="2.0")
    def update_permissions(self, item, rules):
        return self._permissions.update(item, rules)

    @api(version="2.0")
    def delete_permission(self, item, rules):
        self._permissions.delete(item, rules)

    @api(version="2.1")
    def populate_workbook_default_permissions(self, item):
        self._default_permissions.populate_default_permissions(item, Resource.Workbook)

    @api(version="2.1")
    def populate_datasource_default_permissions(self, item):
        self._default_permissions.populate_default_permissions(item, Resource.Datasource)

    @api(version="3.2")
    def populate_metric_default_permissions(self, item):
        self._default_permissions.populate_default_permissions(item, Resource.Metric)

    @api(version="3.4")
    def populate_datarole_default_permissions(self, item):
        self._default_permissions.populate_default_permissions(item, Resource.Datarole)

    @api(version="3.4")
    def populate_flow_default_permissions(self, item):
        self._default_permissions.populate_default_permissions(item, Resource.Flow)

    @api(version="3.4")
    def populate_lens_default_permissions(self, item):
        self._default_permissions.populate_default_permissions(item, Resource.Lens)

    @api(version="2.1")
    def update_workbook_default_permissions(self, item, rules):
        return self._default_permissions.update_default_permissions(item, rules, Resource.Workbook)

    @api(version="2.1")
    def update_datasource_default_permissions(self, item, rules):
        return self._default_permissions.update_default_permissions(item, rules, Resource.Datasource)

    @api(version="3.2")
    def update_metric_default_permissions(self, item, rules):
        return self._default_permissions.update_default_permissions(item, rules, Resource.Metric)

    @api(version="3.4")
    def update_datarole_default_permissions(self, item, rules):
        return self._default_permissions.update_default_permissions(item, rules, Resource.Datarole)

    @api(version="3.4")
    def update_flow_default_permissions(self, item, rules):
        return self._default_permissions.update_default_permissions(item, rules, Resource.Flow)

    @api(version="3.4")
    def update_lens_default_permissions(self, item, rules):
        return self._default_permissions.update_default_permissions(item, rules, Resource.Lens)

    @api(version="2.1")
    def delete_workbook_default_permissions(self, item, rule):
        self._default_permissions.delete_default_permission(item, rule, Resource.Workbook)

    @api(version="2.1")
    def delete_datasource_default_permissions(self, item, rule):
        self._default_permissions.delete_default_permission(item, rule, Resource.Datasource)

    @api(version="3.2")
    def delete_metric_default_permissions(self, item, rule):
        self._default_permissions.delete_default_permission(item, rule, Resource.Metric)

    @api(version="3.4")
    def delete_datarole_default_permissions(self, item, rule):
        self._default_permissions.delete_default_permission(item, rule, Resource.Datarole)

    @api(version="3.4")
    def delete_flow_default_permissions(self, item, rule):
        self._default_permissions.delete_default_permission(item, rule, Resource.Flow)

    @api(version="3.4")
    def delete_lens_default_permissions(self, item, rule):
        self._default_permissions.delete_default_permission(item, rule, Resource.Lens)



import os
import unittest
from io import BytesIO
from xml.etree import ElementTree as ET
from zipfile import ZipFile

from tableauserverclient.filesys_helpers import get_file_object_size, get_file_type
from ._utils import asset, TEST_ASSET_DIR


class FilesysTests(unittest.TestCase):
    def test_get_file_size_returns_correct_size(self):
        target_size = 1000  # bytes

        with BytesIO() as f:
            f.seek(target_size - 1)
            f.write(b"\0")
            file_size = get_file_object_size(f)

        self.assertEqual(file_size, target_size)

    def test_get_file_size_returns_zero_for_empty_file(self):
        with BytesIO() as f:
            file_size = get_file_object_size(f)

        self.assertEqual(file_size, 0)

    def test_get_file_size_coincides_with_built_in_method(self):
        asset_path = asset("SampleWB.twbx")
        target_size = os.path.getsize(asset_path)
        with open(asset_path, "rb") as f:
            file_size = get_file_object_size(f)

        self.assertEqual(file_size, target_size)

    def test_get_file_type_identifies_a_zip_file(self):
        with BytesIO() as file_object:
            with ZipFile(file_object, "w") as zf:
                with BytesIO() as stream:
                    stream.write("This is a zip file".encode())
                    zf.writestr("dummy_file", stream.getbuffer())
                    file_object.seek(0)
                    file_type = get_file_type(file_object)

        self.assertEqual(file_type, "zip")

    def test_get_file_type_identifies_tdsx_as_zip_file(self):
        with open(asset("World Indicators.tdsx"), "rb") as file_object:
            file_type = get_file_type(file_object)
        self.assertEqual(file_type, "zip")

    def test_get_file_type_identifies_twbx_as_zip_file(self):
        with open(asset("SampleWB.twbx"), "rb") as file_object:
            file_type = get_file_type(file_object)
        self.assertEqual(file_type, "zip")

    def test_get_file_type_identifies_xml_file(self):
        root = ET.Element("root")
        child = ET.SubElement(root, "child")
        child.text = "This is a child element"
        etree = ET.ElementTree(root)

        with BytesIO() as file_object:
            etree.write(file_object, encoding="utf-8", xml_declaration=True)

            file_object.seek(0)
            file_type = get_file_type(file_object)

        self.assertEqual(file_type, "xml")

    def test_get_file_type_identifies_tds_as_xml_file(self):
        with open(asset("World Indicators.tds"), "rb") as file_object:
            file_type = get_file_type(file_object)
        self.assertEqual(file_type, "xml")

    def test_get_file_type_identifies_twb_as_xml_file(self):
        with open(asset("RESTAPISample.twb"), "rb") as file_object:
            file_type = get_file_type(file_object)
        self.assertEqual(file_type, "xml")

    def test_get_file_type_identifies_hyper_file(self):
        with open(asset("World Indicators.hyper"), "rb") as file_object:
            file_type = get_file_type(file_object)
        self.assertEqual(file_type, "hyper")

    def test_get_file_type_identifies_tde_file(self):
        asset_path = os.path.join(TEST_ASSET_DIR, "Data", "Tableau Samples", "World Indicators.tde")
        with open(asset_path, "rb") as file_object:
            file_type = get_file_type(file_object)
        self.assertEqual(file_type, "tde")

    def test_get_file_type_handles_unknown_file_type(self):
        # Create a dummy png file
        with BytesIO() as file_object:
            png_signature = bytes.fromhex("89504E470D0A1A0A")
            file_object.write(png_signature)
            file_object.seek(0)

            self.assertRaises(ValueError, get_file_type, file_object)



from functools import partial

from . import RequestOptions


class Pager(object):
    """
    Generator that takes an endpoint (top level endpoints with `.get)` and lazily loads items from Server.
    Supports all `RequestOptions` including starting on any page. Also used by models to load sub-models
    (users in a group, views in a workbook, etc) by passing a different endpoint.

    Will loop over anything that returns (List[ModelItem], PaginationItem).
    """

    def __init__(self, endpoint, request_opts=None, **kwargs):
        if hasattr(endpoint, "get"):
            # The simpliest case is to take an Endpoint and call its get
            endpoint = partial(endpoint.get, **kwargs)
            self._endpoint = endpoint
        elif callable(endpoint):
            # but if they pass a callable then use that instead (used internally)
            endpoint = partial(endpoint, **kwargs)
            self._endpoint = endpoint
        else:
            # Didn't get something we can page over
            raise ValueError("Pager needs a server endpoint to page through.")

        self._options = request_opts

        # If we have options we could be starting on any page, backfill the count
        if self._options:
            self._count = (self._options.pagenumber - 1) * self._options.pagesize
        else:
            self._count = 0
            self._options = RequestOptions()

    def __iter__(self):
        # Fetch the first page
        current_item_list, last_pagination_item = self._endpoint(self._options)

        if last_pagination_item.total_available is None:
            # This endpoint does not support pagination, drain the list and return
            while current_item_list:
                yield current_item_list.pop(0)

            return

        # Get the rest on demand as a generator
        while self._count < last_pagination_item.total_available:
            if len(current_item_list) == 0:
                current_item_list, last_pagination_item = self._load_next_page(last_pagination_item)

            try:
                yield current_item_list.pop(0)
                self._count += 1

            except IndexError:
                # The total count on Server changed while fetching exit gracefully
                return

    def _load_next_page(self, last_pagination_item):
        next_page = last_pagination_item.page_number + 1
        opts = RequestOptions(pagenumber=next_page, pagesize=last_pagination_item.page_size)
        if self._options is not None:
            opts.sort, opts.filter = self._options.sort, self._options.filter
        current_item_list, last_pagination_item = self._endpoint(opts)
        return current_item_list, last_pagination_item



import unittest

try:
    from unittest import mock
except ImportError:
    import mock  # type: ignore[no-redef]

import tableauserverclient.server.request_factory as factory
from tableauserverclient.helpers.strings import redact_xml
from tableauserverclient.filesys_helpers import to_filename, make_download_path


class BugFix257(unittest.TestCase):
    def test_empty_request_works(self):
        result = factory.EmptyRequest().empty_req()
        self.assertEqual(b"<tsRequest />", result)


class FileSysHelpers(unittest.TestCase):
    def test_to_filename(self):
        invalid = [
            "23brhafbjrjhkbbea.txt",
            "a_b_C.txt",
            "windows space.txt",
            "abc#def.txt",
            "t@bL3A()",
        ]

        valid = [
            "23brhafbjrjhkbbea.txt",
            "a_b_C.txt",
            "windows space.txt",
            "abcdef.txt",
            "tbL3A",
        ]

        self.assertTrue(all([(to_filename(i) == v) for i, v in zip(invalid, valid)]))

    def test_make_download_path(self):
        no_file_path = (None, "file.ext")
        has_file_path_folder = ("/root/folder/", "file.ext")
        has_file_path_file = ("outx", "file.ext")

        self.assertEqual("file.ext", make_download_path(*no_file_path))
        self.assertEqual("outx.ext", make_download_path(*has_file_path_file))

        with mock.patch("os.path.isdir") as mocked_isdir:
            mocked_isdir.return_value = True
            self.assertEqual("/root/folder/file.ext", make_download_path(*has_file_path_folder))


class LoggingTest(unittest.TestCase):
    def test_redact_password_string(self):
        redacted = redact_xml(
            "<?xml version='1.0'?><workbook><password>this is password: my_super_secret_passphrase_which_nobody_should_ever_see  password: value</password></workbook>"
        )
        assert redacted.find("value") == -1
        assert redacted.find("secret") == -1
        assert redacted.find("ever_see") == -1
        assert redacted.find("my_super_secret_passphrase_which_nobody_should_ever_see") == -1

    def test_redact_password_bytes(self):
        redacted = redact_xml(
            b"<?xml version='1.0'?><datasource><data-connection name='con-artist' password='value string with at least a password: valuesecret or two in it'/></datasource>"
        )
        assert redacted.find(b"value") == -1
        assert redacted.find(b"secret") == -1

    def test_redact_password_with_special_char(self):
        redacted = redact_xml(
            "<?xml version='1.0'?><content><safe_text value='this is a nondescript text line which is public' password='my_s per_secre>_passphrase_which_nobody_should_ever_see with password: value'> </safe_text></content>"
        )
        assert redacted.find("my_s per_secre>_passphrase_which_nobody_should_ever_see with password: value") == -1

    def test_redact_password_not_xml(self):
        redacted = redact_xml(
            "<content><safe_text value='this is a nondescript text line which is public' password='my_s per_secre>_passphrase_which_nobody_should_ever_see with password: value'> </safe_text></content>"
        )
        assert redacted.find("my_s per_secre>_passphrase_which_nobody_should_ever_see") == -1

    def test_redact_password_really_not_xml(self):
        redacted = redact_xml(
            "value='this is a nondescript text line which is public' password='my_s per_secre>_passphrase_which_nobody_should_ever_see with password: value and then a cookie "
        )
        assert redacted.find("my_s per_secre>_passphrase_which_nobody_should_ever_see") == -1
        assert redacted.find("passphrase") == -1, redacted
        assert redacted.find("cookie") == -1, redacted



import tableauserverclient as TSC
import unittest
import requests
import requests_mock

from unittest import mock
from requests.exceptions import MissingSchema


# This method will be used by the mock to replace requests.get
def mocked_requests_get(*args, **kwargs):
    class MockResponse:
        def __init__(self, status_code):
            self.headers = {}
            self.encoding = None
            self.content = (
                "<xml>"
                "<version version='0.31'>"
                "<api_version>0.31</api_version>"
                "<server_api_version>0.31</server_api_version>"
                "<product_version>2022.3</product_version>"
                "</version>"
                "</xml>"
            )
            self.status_code = status_code

    return MockResponse(200)


class ServerTests(unittest.TestCase):
    def test_init_server_model_empty_throws(self):
        with self.assertRaises(TypeError):
            server = TSC.Server()

    def test_init_server_model_no_protocol_defaults_htt(self):
        server = TSC.Server("fake-url")

    def test_init_server_model_valid_server_name_works(self):
        server = TSC.Server("http://fake-url")

    def test_init_server_model_valid_https_server_name_works(self):
        # by default, it will just set the version to 2.3
        server = TSC.Server("https://fake-url")

    def test_init_server_model_bad_server_name_not_version_check(self):
        server = TSC.Server("fake-url", use_server_version=False)

    @mock.patch("requests.sessions.Session.get", side_effect=mocked_requests_get)
    def test_init_server_model_bad_server_name_do_version_check(self, mock_get):
        server = TSC.Server("fake-url", use_server_version=True)

    def test_init_server_model_bad_server_name_not_version_check_random_options(self):
        # with self.assertRaises(MissingSchema):
        server = TSC.Server("fake-url", use_server_version=False, http_options={"foo": 1})

    def test_init_server_model_bad_server_name_not_version_check_real_options(self):
        # with self.assertRaises(ValueError):
        server = TSC.Server("fake-url", use_server_version=False, http_options={"verify": False})

    def test_http_options_skip_ssl_works(self):
        http_options = {"verify": False}
        server = TSC.Server("http://fake-url")
        server.add_http_options(http_options)

    def test_http_options_multiple_options_works(self):
        http_options = {"verify": False, "birdname": "Parrot"}
        server = TSC.Server("http://fake-url")
        server.add_http_options(http_options)

    # ValueError: dictionary update sequence element #0 has length 1; 2 is required
    def test_http_options_multiple_dicts_fails(self):
        http_options_1 = {"verify": False}
        http_options_2 = {"birdname": "Parrot"}
        server = TSC.Server("http://fake-url")
        with self.assertRaises(ValueError):
            server.add_http_options([http_options_1, http_options_2])

    # TypeError: cannot convert dictionary update sequence element #0 to a sequence
    def test_http_options_not_sequence_fails(self):
        server = TSC.Server("http://fake-url")
        with self.assertRaises(ValueError):
            server.add_http_options({1, 2, 3})

    def test_validate_connection_http(self):
        url = "http://cookies.com"
        server = TSC.Server(url)
        server.validate_connection_settings()
        self.assertEqual(url, server.server_address)

    def test_validate_connection_https(self):
        url = "https://cookies.com"
        server = TSC.Server(url)
        server.validate_connection_settings()
        self.assertEqual(url, server.server_address)

    def test_validate_connection_no_protocol(self):
        url = "cookies.com"
        fixed_url = "http://cookies.com"
        server = TSC.Server(url)
        server.validate_connection_settings()
        self.assertEqual(fixed_url, server.server_address)


class SessionTests(unittest.TestCase):
    test_header = {"x-test": "true"}

    @staticmethod
    def session_factory():
        session = requests.session()
        session.headers.update(SessionTests.test_header)
        return session

    def test_session_factory_adds_headers(self):
        test_request_bin = "http://capture-this-with-mock.com"
        with requests_mock.mock() as m:
            m.get(url="http://capture-this-with-mock.com/api/2.4/serverInfo", request_headers=SessionTests.test_header)
            server = TSC.Server(test_request_bin, use_server_version=True, session_factory=SessionTests.session_factory)



import re
import unittest

import requests
import requests_mock

import tableauserverclient as TSC
from tableauserverclient.server.endpoint.exceptions import InternalServerError, NonXMLResponseError


class RequestTests(unittest.TestCase):
    def setUp(self):
        self.server = TSC.Server("http://test", False)

        # Fake sign in
        self.server._site_id = "dad65087-b08b-4603-af4e-2887b8aafc67"
        self.server._auth_token = "j80k54ll2lfMZ0tv97mlPvvSCRyD0DOM"

        self.baseurl = self.server.workbooks.baseurl

    def test_make_get_request(self):
        with requests_mock.mock() as m:
            m.get(requests_mock.ANY)
            url = "http://test/api/2.3/sites/dad65087-b08b-4603-af4e-2887b8aafc67/workbooks"
            opts = TSC.RequestOptions(pagesize=13, pagenumber=15)
            resp = self.server.workbooks.get_request(url, request_object=opts)

            self.assertTrue(re.search("pagesize=13", resp.request.query))
            self.assertTrue(re.search("pagenumber=15", resp.request.query))

    def test_make_post_request(self):
        with requests_mock.mock() as m:
            m.post(requests_mock.ANY)
            url = "http://test/api/2.3/sites/dad65087-b08b-4603-af4e-2887b8aafc67/workbooks"
            resp = self.server.workbooks._make_request(
                requests.post,
                url,
                content=b"1337",
                auth_token="j80k54ll2lfMZ0tv97mlPvvSCRyD0DOM",
                content_type="multipart/mixed",
            )
            self.assertEqual(resp.request.headers["x-tableau-auth"], "j80k54ll2lfMZ0tv97mlPvvSCRyD0DOM")
            self.assertEqual(resp.request.headers["content-type"], "multipart/mixed")
            self.assertTrue(re.search("Tableau Server Client", resp.request.headers["user-agent"]))
            self.assertEqual(resp.request.body, b"1337")

    # Test that 500 server errors are handled properly
    def test_internal_server_error(self):
        self.server.version = "3.2"
        server_response = "500: Internal Server Error"
        with requests_mock.mock() as m:
            m.register_uri("GET", self.server.server_info.baseurl, status_code=500, text=server_response)
            self.assertRaisesRegex(InternalServerError, server_response, self.server.server_info.get)

    # Test that non-xml server errors are handled properly
    def test_non_xml_error(self):
        self.server.version = "3.2"
        server_response = "this is not xml"
        with requests_mock.mock() as m:
            m.register_uri("GET", self.server.server_info.baseurl, status_code=499, text=server_response)
            self.assertRaisesRegex(NonXMLResponseError, server_response, self.server.server_info.get)



####
# This script demonstrates how to list all of the workbooks or datasources
#
# To run the script, you must have installed Python 3.7 or later.
####

import argparse
import logging
import os
import sys

import tableauserverclient as TSC


def main():
    parser = argparse.ArgumentParser(description="List out the names and LUIDs for different resource types.")
    # Common options; please keep those in sync across all samples
    parser.add_argument("--server", "-s", required=True, help="server address")
    parser.add_argument("--site", "-S", help="site name")
    parser.add_argument(
        "--token-name", "-n", required=True, help="name of the personal access token used to sign into the server"
    )
    parser.add_argument(
        "--token-value", "-v", required=True, help="value of the personal access token used to sign into the server"
    )
    parser.add_argument(
        "--logging-level",
        "-l",
        choices=["debug", "info", "error"],
        default="error",
        help="desired logging level (set to error by default)",
    )
    # Options specific to this sample
    parser.add_argument("resource_type", choices=["workbook", "datasource", "project", "view", "job", "webhooks"])

    args = parser.parse_args()

    # Set logging level based on user input, or error by default
    logging_level = getattr(logging, args.logging_level.upper())
    logging.basicConfig(level=logging_level)

    # Sign in to server
    tableau_auth = TSC.PersonalAccessTokenAuth(args.token_name, args.token_value, site_id=args.site)
    server = TSC.Server(args.server, use_server_version=True)
    with server.auth.sign_in(tableau_auth):
        endpoint = {
            "datasource": server.datasources,
            "job": server.jobs,
            "metric": server.metrics,
            "project": server.projects,
            "view": server.views,
            "webhooks": server.webhooks,
            "workbook": server.workbooks,
        }.get(args.resource_type)

        options = TSC.RequestOptions()
        options.sort.add(TSC.Sort(TSC.RequestOptions.Field.Name, TSC.RequestOptions.Direction.Desc))

        count = 0
        for resource in TSC.Pager(endpoint.get, options):
            count = count + 1
            # endpoint.populate_connections(resource)
            print(resource.name[:18], " ")  # , resource._connections())
            if count > 100:
                break
        print("Total: {}".format(count))


if __name__ == "__main__":
    main()



####
# This script demonstrates how to export a view using the Tableau
# Server Client.
#
# To run the script, you must have installed Python 3.7 or later.
####

import argparse
import logging

import tableauserverclient as TSC


def main():
    parser = argparse.ArgumentParser(description="Export a view as an image, PDF, or CSV")
    # Common options; please keep those in sync across all samples
    parser.add_argument("--server", "-s", required=True, help="server address")
    parser.add_argument("--site", "-S", help="site name")
    parser.add_argument(
        "--token-name", "-p", required=True, help="name of the personal access token used to sign into the server"
    )
    parser.add_argument(
        "--token-value", "-v", required=True, help="value of the personal access token used to sign into the server"
    )
    parser.add_argument(
        "--logging-level",
        "-l",
        choices=["debug", "info", "error"],
        default="error",
        help="desired logging level (set to error by default)",
    )
    # Options specific to this sample
    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument(
        "--pdf", dest="type", action="store_const", const=("populate_pdf", "PDFRequestOptions", "pdf", "pdf")
    )
    group.add_argument(
        "--png", dest="type", action="store_const", const=("populate_image", "ImageRequestOptions", "image", "png")
    )
    group.add_argument(
        "--csv", dest="type", action="store_const", const=("populate_csv", "CSVRequestOptions", "csv", "csv")
    )
    # other options shown in explore_workbooks: workbook.download, workbook.preview_image

    parser.add_argument("--workbook", action="store_true")

    parser.add_argument("--file", "-f", help="filename to store the exported data")
    parser.add_argument("--filter", "-vf", metavar="COLUMN:VALUE", help="View filter to apply to the view")
    parser.add_argument("resource_id", help="LUID for the view or workbook")

    args = parser.parse_args()

    # Set logging level based on user input, or error by default
    logging_level = getattr(logging, args.logging_level.upper())
    logging.basicConfig(level=logging_level)

    tableau_auth = TSC.PersonalAccessTokenAuth(args.token_name, args.token_value, site_id=args.site)
    server = TSC.Server(args.server, use_server_version=True, http_options={"verify": False})
    with server.auth.sign_in(tableau_auth):
        print("Connected")
        if args.workbook:
            item = server.workbooks.get_by_id(args.resource_id)
        else:
            item = server.views.get_by_id(args.resource_id)

        if not item:
            print("No item found for id {}".format(args.resource_id))
            exit(1)

        print("Item found: {}".format(item.name))
        # We have a number of different types and functions for each different export type.
        # We encode that information above in the const=(...) parameter to the add_argument function to make
        # the code automatically adapt for the type of export the user is doing.
        # We unroll that information into methods we can call, or objects we can create by using getattr()
        (populate_func_name, option_factory_name, member_name, extension) = args.type
        populate = getattr(server.views, populate_func_name)
        if args.workbook:
            populate = getattr(server.workbooks, populate_func_name)

        option_factory = getattr(TSC, option_factory_name)

        if args.filter:
            options = option_factory().vf(*args.filter.split(":"))
        else:
            options = None

        if args.file:
            filename = args.file
        else:
            filename = "out.{}".format(extension)

        populate(item, options)
        with open(filename, "wb") as f:
            if member_name == "csv":
                f.writelines(getattr(item, member_name))
            else:
                f.write(getattr(item, member_name))
        print("saved to " + filename)


if __name__ == "__main__":
    main()



import os
from pathlib import Path
import re
import unittest

import requests_mock

import tableauserverclient as TSC

TEST_ASSET_DIR = Path(__file__).parent / "assets"

PAGINATION_XML = os.path.join(TEST_ASSET_DIR, "request_option_pagination.xml")
PAGE_NUMBER_XML = os.path.join(TEST_ASSET_DIR, "request_option_page_number.xml")
PAGE_SIZE_XML = os.path.join(TEST_ASSET_DIR, "request_option_page_size.xml")
FILTER_EQUALS = os.path.join(TEST_ASSET_DIR, "request_option_filter_equals.xml")
FILTER_TAGS_IN = os.path.join(TEST_ASSET_DIR, "request_option_filter_tags_in.xml")
FILTER_MULTIPLE = os.path.join(TEST_ASSET_DIR, "request_option_filter_tags_in.xml")
SLICING_QUERYSET = os.path.join(TEST_ASSET_DIR, "request_option_slicing_queryset.xml")
SLICING_QUERYSET_PAGE_1 = TEST_ASSET_DIR / "queryset_slicing_page_1.xml"
SLICING_QUERYSET_PAGE_2 = TEST_ASSET_DIR / "queryset_slicing_page_2.xml"


class RequestOptionTests(unittest.TestCase):
    def setUp(self) -> None:
        self.server = TSC.Server("http://test", False)

        # Fake signin
        self.server.version = "3.10"
        self.server._site_id = "dad65087-b08b-4603-af4e-2887b8aafc67"
        self.server._auth_token = "j80k54ll2lfMZ0tv97mlPvvSCRyD0DOM"

        self.baseurl = "{0}/{1}".format(self.server.sites.baseurl, self.server._site_id)

    def test_pagination(self) -> None:
        with open(PAGINATION_XML, "rb") as f:
            response_xml = f.read().decode("utf-8")
        with requests_mock.mock() as m:
            m.get(self.baseurl + "/views?pageNumber=1&pageSize=10", text=response_xml)
            req_option = TSC.RequestOptions().page_size(10)
            all_views, pagination_item = self.server.views.get(req_option)

        self.assertEqual(1, pagination_item.page_number)
        self.assertEqual(10, pagination_item.page_size)
        self.assertEqual(33, pagination_item.total_available)
        self.assertEqual(10, len(all_views))

    def test_page_number(self) -> None:
        with open(PAGE_NUMBER_XML, "rb") as f:
            response_xml = f.read().decode("utf-8")
        with requests_mock.mock() as m:
            m.get(self.baseurl + "/views?pageNumber=3", text=response_xml)
            req_option = TSC.RequestOptions().page_number(3)
            all_views, pagination_item = self.server.views.get(req_option)

        self.assertEqual(3, pagination_item.page_number)
        self.assertEqual(100, pagination_item.page_size)
        self.assertEqual(210, pagination_item.total_available)
        self.assertEqual(10, len(all_views))

    def test_page_size(self) -> None:
        with open(PAGE_SIZE_XML, "rb") as f:
            response_xml = f.read().decode("utf-8")
        with requests_mock.mock() as m:
            m.get(self.baseurl + "/views?pageSize=5", text=response_xml)
            req_option = TSC.RequestOptions().page_size(5)
            all_views, pagination_item = self.server.views.get(req_option)

        self.assertEqual(1, pagination_item.page_number)
        self.assertEqual(5, pagination_item.page_size)
        self.assertEqual(33, pagination_item.total_available)
        self.assertEqual(5, len(all_views))

    def test_filter_equals(self) -> None:
        with open(FILTER_EQUALS, "rb") as f:
            response_xml = f.read().decode("utf-8")
        with requests_mock.mock() as m:
            m.get(self.baseurl + "/workbooks?filter=name:eq:RESTAPISample", text=response_xml)
            req_option = TSC.RequestOptions()
            req_option.filter.add(
                TSC.Filter(TSC.RequestOptions.Field.Name, TSC.RequestOptions.Operator.Equals, "RESTAPISample")
            )
            matching_workbooks, pagination_item = self.server.workbooks.get(req_option)

        self.assertEqual(2, pagination_item.total_available)
        self.assertEqual("RESTAPISample", matching_workbooks[0].name)
        self.assertEqual("RESTAPISample", matching_workbooks[1].name)

    def test_filter_equals_shorthand(self) -> None:
        with open(FILTER_EQUALS, "rb") as f:
            response_xml = f.read().decode("utf-8")
        with requests_mock.mock() as m:
            m.get(self.baseurl + "/workbooks?filter=name:eq:RESTAPISample", text=response_xml)
            matching_workbooks = self.server.workbooks.filter(name="RESTAPISample").order_by("name")

            self.assertEqual(2, matching_workbooks.total_available)
            self.assertEqual("RESTAPISample", matching_workbooks[0].name)
            self.assertEqual("RESTAPISample", matching_workbooks[1].name)

    def test_filter_tags_in(self) -> None:
        with open(FILTER_TAGS_IN, "rb") as f:
            response_xml = f.read().decode("utf-8")
        with requests_mock.mock() as m:
            m.get(self.baseurl + "/workbooks?filter=tags:in:[sample,safari,weather]", text=response_xml)
            req_option = TSC.RequestOptions()
            req_option.filter.add(
                TSC.Filter(
                    TSC.RequestOptions.Field.Tags, TSC.RequestOptions.Operator.In, ["sample", "safari", "weather"]
                )
            )
            matching_workbooks, pagination_item = self.server.workbooks.get(req_option)

        self.assertEqual(3, pagination_item.total_available)
        self.assertEqual(set(["weather"]), matching_workbooks[0].tags)
        self.assertEqual(set(["safari"]), matching_workbooks[1].tags)
        self.assertEqual(set(["sample"]), matching_workbooks[2].tags)

    def test_filter_tags_in_shorthand(self) -> None:
        with open(FILTER_TAGS_IN, "rb") as f:
            response_xml = f.read().decode("utf-8")
        with requests_mock.mock() as m:
            m.get(self.baseurl + "/workbooks?filter=tags:in:[sample,safari,weather]", text=response_xml)
            matching_workbooks = self.server.workbooks.filter(tags__in=["sample", "safari", "weather"])

            self.assertEqual(3, matching_workbooks.total_available)
            self.assertEqual(set(["weather"]), matching_workbooks[0].tags)
            self.assertEqual(set(["safari"]), matching_workbooks[1].tags)
            self.assertEqual(set(["sample"]), matching_workbooks[2].tags)

    def test_invalid_shorthand_option(self) -> None:
        with self.assertRaises(ValueError):
            self.server.workbooks.filter(nonexistant__in=["sample", "safari"])

    def test_multiple_filter_options(self) -> None:
        with open(FILTER_MULTIPLE, "rb") as f:
            response_xml = f.read().decode("utf-8")
        # To ensure that this is deterministic, run this a few times
        with requests_mock.mock() as m:
            # Sometimes pep8 requires you to do things you might not otherwise do
            url = "".join(
                (
                    self.baseurl,
                    "/workbooks?pageNumber=1&pageSize=100&",
                    "filter=name:eq:foo,tags:in:[sample,safari,weather]",
                )
            )
            m.get(url, text=response_xml)
            req_option = TSC.RequestOptions()
            req_option.filter.add(
                TSC.Filter(
                    TSC.RequestOptions.Field.Tags, TSC.RequestOptions.Operator.In, ["sample", "safari", "weather"]
                )
            )
            req_option.filter.add(TSC.Filter(TSC.RequestOptions.Field.Name, TSC.RequestOptions.Operator.Equals, "foo"))
            for _ in range(100):
                matching_workbooks, pagination_item = self.server.workbooks.get(req_option)
                self.assertEqual(3, pagination_item.total_available)

    # Test req_options if url already has query params
    def test_double_query_params(self) -> None:
        with requests_mock.mock() as m:
            m.get(requests_mock.ANY)
            url = self.baseurl + "/views?queryParamExists=true"
            opts = TSC.RequestOptions()

            opts.filter.add(
                TSC.Filter(TSC.RequestOptions.Field.Tags, TSC.RequestOptions.Operator.In, ["stocks", "market"])
            )
            opts.sort.add(TSC.Sort(TSC.RequestOptions.Field.Name, TSC.RequestOptions.Direction.Asc))

            resp = self.server.workbooks.get_request(url, request_object=opts)
            self.assertTrue(re.search("queryparamexists=true", resp.request.query))
            self.assertTrue(re.search("filter=tags%3ain%3a%5bstocks%2cmarket%5d", resp.request.query))
            self.assertTrue(re.search("sort=name%3aasc", resp.request.query))

    # Test req_options for versions below 3.7
    def test_filter_sort_legacy(self) -> None:
        self.server.version = "3.6"
        with requests_mock.mock() as m:
            m.get(requests_mock.ANY)
            url = self.baseurl + "/views?queryParamExists=true"
            opts = TSC.RequestOptions()

            opts.filter.add(
                TSC.Filter(TSC.RequestOptions.Field.Tags, TSC.RequestOptions.Operator.In, ["stocks", "market"])
            )
            opts.sort.add(TSC.Sort(TSC.RequestOptions.Field.Name, TSC.RequestOptions.Direction.Asc))

            resp = self.server.workbooks.get_request(url, request_object=opts)
            self.assertTrue(re.search("queryparamexists=true", resp.request.query))
            self.assertTrue(re.search("filter=tags:in:%5bstocks,market%5d", resp.request.query))
            self.assertTrue(re.search("sort=name:asc", resp.request.query))

    def test_vf(self) -> None:
        with requests_mock.mock() as m:
            m.get(requests_mock.ANY)
            url = self.baseurl + "/views/456/data"
            opts = TSC.PDFRequestOptions()
            opts.vf("name1#", "value1")
            opts.vf("name2$", "value2")
            opts.page_type = TSC.PDFRequestOptions.PageType.Tabloid

            resp = self.server.workbooks.get_request(url, request_object=opts)
            self.assertTrue(re.search("vf_name1%23=value1", resp.request.query))
            self.assertTrue(re.search("vf_name2%24=value2", resp.request.query))
            self.assertTrue(re.search("type=tabloid", resp.request.query))

    # Test req_options for versions beloe 3.7
    def test_vf_legacy(self) -> None:
        self.server.version = "3.6"
        with requests_mock.mock() as m:
            m.get(requests_mock.ANY)
            url = self.baseurl + "/views/456/data"
            opts = TSC.PDFRequestOptions()
            opts.vf("name1@", "value1")
            opts.vf("name2$", "value2")
            opts.page_type = TSC.PDFRequestOptions.PageType.Tabloid

            resp = self.server.workbooks.get_request(url, request_object=opts)
            self.assertTrue(re.search("vf_name1@=value1", resp.request.query))
            self.assertTrue(re.search("vf_name2\\$=value2", resp.request.query))
            self.assertTrue(re.search("type=tabloid", resp.request.query))

    def test_all_fields(self) -> None:
        with requests_mock.mock() as m:
            m.get(requests_mock.ANY)
            url = self.baseurl + "/views/456/data"
            opts = TSC.RequestOptions()
            opts._all_fields = True

            resp = self.server.users.get_request(url, request_object=opts)
            self.assertTrue(re.search("fields=_all_", resp.request.query))

    def test_multiple_filter_options_shorthand(self) -> None:
        with open(FILTER_MULTIPLE, "rb") as f:
            response_xml = f.read().decode("utf-8")
        # To ensure that this is deterministic, run this a few times
        with requests_mock.mock() as m:
            # Sometimes pep8 requires you to do things you might not otherwise do
            url = "".join(
                (
                    self.baseurl,
                    "/workbooks?pageNumber=1&pageSize=100&",
                    "filter=name:eq:foo,tags:in:[sample,safari,weather]",
                )
            )
            m.get(url, text=response_xml)

            for _ in range(100):
                matching_workbooks = self.server.workbooks.filter(tags__in=["sample", "safari", "weather"], name="foo")
                self.assertEqual(3, matching_workbooks.total_available)

    def test_slicing_queryset(self) -> None:
        with open(SLICING_QUERYSET, "rb") as f:
            response_xml = f.read().decode("utf-8")
        with requests_mock.mock() as m:
            m.get(self.baseurl + "/views?pageNumber=1", text=response_xml)
            all_views = self.server.views.all()

            self.assertEqual(10, len(all_views[::]))
            self.assertEqual(5, len(all_views[::2]))
            self.assertEqual(8, len(all_views[2:]))
            self.assertEqual(2, len(all_views[:2]))
            self.assertEqual(3, len(all_views[2:5]))
            self.assertEqual(3, len(all_views[-3:]))
            self.assertEqual(3, len(all_views[-6:-3]))
            self.assertEqual(3, len(all_views[3:6:-1]))
            self.assertEqual(3, len(all_views[6:3:-1]))
            self.assertEqual(10, len(all_views[::-1]))
            self.assertEqual(all_views[3:6], list(reversed(all_views[3:6:-1])))

            self.assertEqual(all_views[-3].id, "2df55de2-3a2d-4e34-b515-6d4e70b830e9")

        with self.assertRaises(IndexError):
            all_views[100]

    def test_slicing_queryset_multi_page(self) -> None:
        with requests_mock.mock() as m:
            m.get(self.baseurl + "/views?pageNumber=1", text=SLICING_QUERYSET_PAGE_1.read_text())
            m.get(self.baseurl + "/views?pageNumber=2", text=SLICING_QUERYSET_PAGE_2.read_text())
            sliced_views = self.server.views.all()[9:12]

        self.assertEqual(sliced_views[0].id, "2e6d6c81-da71-4b41-892c-ba80d4e7a6d0")
        self.assertEqual(sliced_views[1].id, "47ffcb8e-3f7a-4ecf-8ab3-605da9febe20")
        self.assertEqual(sliced_views[2].id, "6757fea8-0aa9-4160-a87c-9be27b1d1c8c")

    def test_queryset_filter_args_error(self) -> None:
        with self.assertRaises(RuntimeError):
            workbooks = self.server.workbooks.filter("argument")



import io
import xml.etree.ElementTree as ET
from datetime import datetime
from enum import IntEnum
from typing import Dict, List, Optional, TYPE_CHECKING, Tuple

from defusedxml.ElementTree import fromstring

from tableauserverclient.datetime_helpers import parse_datetime
from .exceptions import UnpopulatedPropertyError
from .property_decorators import (
    property_is_enum,
    property_not_empty,
)
from .reference_item import ResourceReference

if TYPE_CHECKING:
    from tableauserverclient.server import Pager


class UserItem(object):
    tag_name: str = "user"

    class Roles:
        Interactor = "Interactor"
        Publisher = "Publisher"
        ServerAdministrator = "ServerAdministrator"
        SiteAdministrator = "SiteAdministrator"
        Unlicensed = "Unlicensed"
        UnlicensedWithPublish = "UnlicensedWithPublish"
        Viewer = "Viewer"
        ViewerWithPublish = "ViewerWithPublish"
        Guest = "Guest"

        Creator = "Creator"
        Explorer = "Explorer"
        ExplorerCanPublish = "ExplorerCanPublish"
        ReadOnly = "ReadOnly"
        SiteAdministratorCreator = "SiteAdministratorCreator"
        SiteAdministratorExplorer = "SiteAdministratorExplorer"

        # Online only
        SupportUser = "SupportUser"

    class Auth:
        OpenID = "OpenID"
        SAML = "SAML"
        ServerDefault = "ServerDefault"

    def __init__(
        self, name: Optional[str] = None, site_role: Optional[str] = None, auth_setting: Optional[str] = None
    ) -> None:
        self._auth_setting: Optional[str] = None
        self._domain_name: Optional[str] = None
        self._external_auth_user_id: Optional[str] = None
        self._id: Optional[str] = None
        self._last_login: Optional[datetime] = None
        self._workbooks = None
        self._favorites: Optional[Dict[str, List]] = None
        self._groups = None
        self.email: Optional[str] = None
        self.fullname: Optional[str] = None
        self.name: Optional[str] = name
        self.site_role: Optional[str] = site_role
        self.auth_setting: Optional[str] = auth_setting

        return None

    def __repr__(self) -> str:
        str_site_role = self.site_role or "None"
        return "<User {} name={} role={}>".format(self.id, self.name, str_site_role)

    @property
    def auth_setting(self) -> Optional[str]:
        return self._auth_setting

    @auth_setting.setter
    @property_is_enum(Auth)
    def auth_setting(self, value):
        self._auth_setting = value

    @property
    def domain_name(self) -> Optional[str]:
        return self._domain_name

    @property
    def external_auth_user_id(self) -> Optional[str]:
        return self._external_auth_user_id

    @property
    def id(self) -> Optional[str]:
        return self._id

    @id.setter
    def id(self, value: str) -> None:
        self._id = value

    @property
    def last_login(self) -> Optional[datetime]:
        return self._last_login

    @property
    def name(self) -> Optional[str]:
        return self._name

    @name.setter
    def name(self, value: str):
        self._name = value

    # valid: username, domain/username, username@domain, domain/username@email
    @staticmethod
    def validate_username_or_throw(username) -> None:
        if username is None or username == "" or username.strip(" ") == "":
            raise AttributeError("Username cannot be empty")
        if username.find(" ") >= 0:
            raise AttributeError("Username cannot contain spaces")
        at_symbol = username.find("@")
        if at_symbol >= 0:
            username = username[:at_symbol] + "X" + username[at_symbol + 1 :]
            if username.find("@") >= 0:
                raise AttributeError("Username cannot repeat '@'")

    @property
    def site_role(self) -> Optional[str]:
        return self._site_role

    @site_role.setter
    @property_is_enum(Roles)
    def site_role(self, value):
        self._site_role = value

    @property
    def workbooks(self) -> "Pager":
        if self._workbooks is None:
            error = "User item must be populated with workbooks first."
            raise UnpopulatedPropertyError(error)
        return self._workbooks()

    @property
    def favorites(self) -> Dict[str, List]:
        if self._favorites is None:
            error = "User item must be populated with favorites first."
            raise UnpopulatedPropertyError(error)
        return self._favorites

    @property
    def groups(self) -> "Pager":
        if self._groups is None:
            error = "User item must be populated with groups first."
            raise UnpopulatedPropertyError(error)
        return self._groups()

    def _set_workbooks(self, workbooks) -> None:
        self._workbooks = workbooks

    def _set_groups(self, groups) -> None:
        self._groups = groups

    def _parse_common_tags(self, user_xml, ns) -> "UserItem":
        if not isinstance(user_xml, ET.Element):
            user_xml = fromstring(user_xml).find(".//t:user", namespaces=ns)
        if user_xml is not None:
            (
                _,
                _,
                site_role,
                _,
                _,
                fullname,
                email,
                auth_setting,
                _,
            ) = self._parse_element(user_xml, ns)
            self._set_values(None, None, site_role, None, None, fullname, email, auth_setting, None)
        return self

    def _set_values(
        self,
        id,
        name,
        site_role,
        last_login,
        external_auth_user_id,
        fullname,
        email,
        auth_setting,
        domain_name,
    ):
        if id is not None:
            self._id = id
        if name:
            self._name = name
        if site_role:
            self._site_role = site_role
        if last_login:
            self._last_login = last_login
        if external_auth_user_id:
            self._external_auth_user_id = external_auth_user_id
        if fullname:
            self.fullname = fullname
        if email:
            self.email = email
        if auth_setting:
            self._auth_setting = auth_setting
        if domain_name:
            self._domain_name = domain_name

    @classmethod
    def from_response(cls, resp, ns) -> List["UserItem"]:
        element_name = ".//t:user"
        return cls._parse_xml(element_name, resp, ns)

    @classmethod
    def from_response_as_owner(cls, resp, ns) -> List["UserItem"]:
        element_name = ".//t:owner"
        return cls._parse_xml(element_name, resp, ns)

    @classmethod
    def _parse_xml(cls, element_name, resp, ns):
        all_user_items = []
        parsed_response = fromstring(resp)
        all_user_xml = parsed_response.findall(element_name, namespaces=ns)
        for user_xml in all_user_xml:
            (
                id,
                name,
                site_role,
                last_login,
                external_auth_user_id,
                fullname,
                email,
                auth_setting,
                domain_name,
            ) = cls._parse_element(user_xml, ns)
            user_item = cls(name, site_role)
            user_item._set_values(
                id,
                name,
                site_role,
                last_login,
                external_auth_user_id,
                fullname,
                email,
                auth_setting,
                domain_name,
            )
            all_user_items.append(user_item)
        return all_user_items

    @staticmethod
    def as_reference(id_) -> ResourceReference:
        return ResourceReference(id_, UserItem.tag_name)

    @staticmethod
    def _parse_element(user_xml, ns):
        id = user_xml.get("id", None)
        name = user_xml.get("name", None)
        site_role = user_xml.get("siteRole", None)
        last_login = parse_datetime(user_xml.get("lastLogin", None))
        external_auth_user_id = user_xml.get("externalAuthUserId", None)
        fullname = user_xml.get("fullName", None)
        email = user_xml.get("email", None)
        auth_setting = user_xml.get("authSetting", None)

        domain_name = None
        domain_elem = user_xml.find(".//t:domain", namespaces=ns)
        if domain_elem is not None:
            domain_name = domain_elem.get("name", None)

        return (
            id,
            name,
            site_role,
            last_login,
            external_auth_user_id,
            fullname,
            email,
            auth_setting,
            domain_name,
        )

    class CSVImport(object):
        """
        This class includes hardcoded options and logic for the CSV file format defined for user import
        https://help.tableau.com/current/server/en-us/users_import.htm
        """

        # username, password, display_name, license, admin_level, publishing, email, auth type
        class ColumnType(IntEnum):
            USERNAME = 0
            PASS = 1
            DISPLAY_NAME = 2
            LICENSE = 3  # aka site role
            ADMIN = 4
            PUBLISHER = 5
            EMAIL = 6
            AUTH = 7

            MAX = 7

        # Read a csv line and create a user item populated by the given attributes
        @staticmethod
        def create_user_from_line(line: str):
            if line is None or line is False or line == "\n" or line == "":
                return None
            line = line.strip().lower()
            values: List[str] = list(map(str.strip, line.split(",")))
            user = UserItem(values[UserItem.CSVImport.ColumnType.USERNAME])
            if len(values) > 1:
                if len(values) > UserItem.CSVImport.ColumnType.MAX:
                    raise ValueError("Too many attributes for user import")
                while len(values) <= UserItem.CSVImport.ColumnType.MAX:
                    values.append("")
                site_role = UserItem.CSVImport._evaluate_site_role(
                    values[UserItem.CSVImport.ColumnType.LICENSE],
                    values[UserItem.CSVImport.ColumnType.ADMIN],
                    values[UserItem.CSVImport.ColumnType.PUBLISHER],
                )

                user._set_values(
                    None,
                    values[UserItem.CSVImport.ColumnType.USERNAME],
                    site_role,
                    None,
                    None,
                    values[UserItem.CSVImport.ColumnType.DISPLAY_NAME],
                    values[UserItem.CSVImport.ColumnType.EMAIL],
                    values[UserItem.CSVImport.ColumnType.AUTH],
                    None,
                )
            return user

        # Read through an entire CSV file meant for user import
        # Return the number of valid lines and a list of all the invalid lines
        @staticmethod
        def validate_file_for_import(csv_file: io.TextIOWrapper, logger) -> Tuple[int, List[str]]:
            num_valid_lines = 0
            invalid_lines = []
            csv_file.seek(0)  # set to start of file in case it has been read earlier
            line: str = csv_file.readline()
            while line and line != "":
                try:
                    # do not print passwords
                    logger.info("Reading user {}".format(line[:4]))
                    UserItem.CSVImport._validate_import_line_or_throw(line, logger)
                    num_valid_lines += 1
                except Exception as exc:
                    logger.info("Error parsing {}: {}".format(line[:4], exc))
                    invalid_lines.append(line)
                line = csv_file.readline()
            return num_valid_lines, invalid_lines

        # Some fields in the import file are restricted to specific values
        # Iterate through each field and validate the given value against hardcoded constraints
        @staticmethod
        def _validate_import_line_or_throw(incoming, logger) -> None:
            _valid_attributes: List[List[str]] = [
                [],
                [],
                [],
                ["creator", "explorer", "viewer", "unlicensed"],  # license
                ["system", "site", "none", "no"],  # admin
                ["yes", "true", "1", "no", "false", "0"],  # publisher
                [],
                [UserItem.Auth.SAML, UserItem.Auth.OpenID, UserItem.Auth.ServerDefault],  # auth
            ]

            line = list(map(str.strip, incoming.split(",")))
            if len(line) > UserItem.CSVImport.ColumnType.MAX:
                raise AttributeError("Too many attributes in line")
            username = line[UserItem.CSVImport.ColumnType.USERNAME.value]
            logger.debug("> details - {}".format(username))
            UserItem.validate_username_or_throw(username)
            for i in range(1, len(line)):
                logger.debug("column {}: {}".format(UserItem.CSVImport.ColumnType(i).name, line[i]))
                UserItem.CSVImport._validate_attribute_value(
                    line[i], _valid_attributes[i], UserItem.CSVImport.ColumnType(i)
                )

        # Given a restricted set of possible values, confirm the item is in that set
        @staticmethod
        def _validate_attribute_value(item: str, possible_values: List[str], column_type) -> None:
            if item is None or item == "":
                # value can be empty for any column except user, which is checked elsewhere
                return
            if item in possible_values or possible_values == []:
                return
            raise AttributeError("Invalid value {} for {}".format(item, column_type))

        # https://help.tableau.com/current/server/en-us/csvguidelines.htm#settings_and_site_roles
        # This logic is hardcoded to match the existing rules for import csv files
        @staticmethod
        def _evaluate_site_role(license_level, admin_level, publisher):
            if not license_level or not admin_level or not publisher:
                return "Unlicensed"
            # ignore case everywhere
            license_level = license_level.lower()
            admin_level = admin_level.lower()
            publisher = publisher.lower()
            # don't need to check publisher for system/site admin
            if admin_level == "system":
                site_role = "SiteAdministrator"
            elif admin_level == "site":
                if license_level == "creator":
                    site_role = "SiteAdministratorCreator"
                elif license_level == "explorer":
                    site_role = "SiteAdministratorExplorer"
                else:
                    site_role = "SiteAdministratorExplorer"
            else:  # if it wasn't 'system' or 'site' then we can treat it as 'none'
                if publisher == "yes":
                    if license_level == "creator":
                        site_role = "Creator"
                    elif license_level == "explorer":
                        site_role = "ExplorerCanPublish"
                    else:
                        site_role = "Unlicensed"  # is this the expected outcome?
                else:  # publisher == 'no':
                    if license_level == "explorer" or license_level == "creator":
                        site_role = "Explorer"
                    elif license_level == "viewer":
                        site_role = "Viewer"
                    else:  # if license_level == 'unlicensed'
                        site_role = "Unlicensed"
            if site_role is None:
                site_role = "Unlicensed"
            return site_role



####
# This script demonstrates how to log in to Tableau Server Client.
#
# To run the script, you must have installed Python 3.7 or later.
####

import argparse
import getpass
import logging

import tableauserverclient as TSC


# If a sample has additional arguments, then it should copy this code and insert them after the call to
# sample_define_common_options
# If it has no additional arguments, it can just call this method
def set_up_and_log_in():
    parser = argparse.ArgumentParser(description="Logs in to the server.")
    sample_define_common_options(parser)
    args = parser.parse_args()

    # Set logging level based on user input, or error by default.
    logging_level = getattr(logging, args.logging_level.upper())
    logging.basicConfig(level=logging_level)

    server = sample_connect_to_server(args)
    print(server.server_info.get())
    print(server.server_address, "site:", server.site_id, "user:", server.user_id)


def sample_define_common_options(parser):
    # Common options; please keep these in sync across all samples by copying or calling this method directly
    parser.add_argument("--server", "-s", required=True, help="server address")
    parser.add_argument("--site", "-t", help="site name")
    auth = parser.add_mutually_exclusive_group(required=True)
    auth.add_argument("--token-name", "-tn", help="name of the personal access token used to sign into the server")
    auth.add_argument("--username", "-u", help="username to sign into the server")

    parser.add_argument("--token-value", "-tv", help="value of the personal access token used to sign into the server")
    parser.add_argument("--password", "-p", help="value of the password used to sign into the server")
    parser.add_argument(
        "--logging-level",
        "-l",
        choices=["debug", "info", "error"],
        default="error",
        help="desired logging level (set to error by default)",
    )


def sample_connect_to_server(args):
    if args.username:
        # Trying to authenticate using username and password.
        password = args.password or getpass.getpass("Password: ")

        tableau_auth = TSC.TableauAuth(args.username, password, site_id=args.site)
        print("\nSigning in...\nServer: {}\nSite: {}\nUsername: {}".format(args.server, args.site, args.username))

    else:
        # Trying to authenticate using personal access tokens.
        token = args.token_value or getpass.getpass("Personal Access Token: ")

        tableau_auth = TSC.PersonalAccessTokenAuth(
            token_name=args.token_name, personal_access_token=token, site_id=args.site
        )
        print("\nSigning in...\nServer: {}\nSite: {}\nToken name: {}".format(args.server, args.site, args.token_name))

    if not tableau_auth:
        raise TabError("Did not create authentication object. Check arguments.")

    # Only set this to False if you are running against a server you trust AND you know why the cert is broken
    check_ssl_certificate = True

    # Make sure we use an updated version of the rest apis, and pass in our cert handling choice
    server = TSC.Server(args.server, use_server_version=True, http_options={"verify": check_ssl_certificate})
    server.auth.sign_in(tableau_auth)
    print("Logged in successfully")

    return server


if __name__ == "__main__":
    set_up_and_log_in()



import logging
import xml.etree.ElementTree as ET
from typing import List, Optional

from defusedxml.ElementTree import fromstring

from .exceptions import UnpopulatedPropertyError
from .property_decorators import property_is_enum, property_not_empty


class ProjectItem(object):
    class ContentPermissions:
        LockedToProject: str = "LockedToProject"
        ManagedByOwner: str = "ManagedByOwner"
        LockedToProjectWithoutNested: str = "LockedToProjectWithoutNested"

    def __repr__(self):
        return "<Project {0} {1} parent={2} permissions={3}>".format(
            self._id, self.name, self.parent_id or "None (Top level)", self.content_permissions or "Not Set"
        )

    def __init__(
        self,
        name: str,
        description: Optional[str] = None,
        content_permissions: Optional[str] = None,
        parent_id: Optional[str] = None,
    ) -> None:
        self._content_permissions = None
        self._id: Optional[str] = None
        self.description: Optional[str] = description
        self.name: str = name
        self.content_permissions: Optional[str] = content_permissions
        self.parent_id: Optional[str] = parent_id

        self._permissions = None
        self._default_workbook_permissions = None
        self._default_datasource_permissions = None
        self._default_flow_permissions = None
        self._default_lens_permissions = None
        self._default_datarole_permissions = None
        self._default_metric_permissions = None

    @property
    def content_permissions(self):
        return self._content_permissions

    @content_permissions.setter
    @property_is_enum(ContentPermissions)
    def content_permissions(self, value: Optional[str]) -> None:
        self._content_permissions = value

    @property
    def permissions(self):
        if self._permissions is None:
            error = "Project item must be populated with permissions first."
            raise UnpopulatedPropertyError(error)
        return self._permissions()

    @property
    def default_datasource_permissions(self):
        if self._default_datasource_permissions is None:
            error = "Project item must be populated with permissions first."
            raise UnpopulatedPropertyError(error)
        return self._default_datasource_permissions()

    @property
    def default_workbook_permissions(self):
        if self._default_workbook_permissions is None:
            error = "Project item must be populated with permissions first."
            raise UnpopulatedPropertyError(error)
        return self._default_workbook_permissions()

    @property
    def default_flow_permissions(self):
        if self._default_flow_permissions is None:
            error = "Project item must be populated with permissions first."
            raise UnpopulatedPropertyError(error)
        return self._default_flow_permissions()

    @property
    def default_lens_permissions(self):
        if self._default_lens_permissions is None:
            error = "Project item must be populated with permissions first."
            raise UnpopulatedPropertyError(error)
        return self._default_lens_permissions()

    @property
    def default_datarole_permissions(self):
        if self._default_datarole_permissions is None:
            error = "Project item must be populated with permissions first."
            raise UnpopulatedPropertyError(error)
        return self._default_datarole_permissions()

    @property
    def default_metric_permissions(self):
        if self._default_metric_permissions is None:
            error = "Project item must be populated with permissions first."
            raise UnpopulatedPropertyError(error)
        return self._default_metric_permissions()

    @property
    def id(self) -> Optional[str]:
        return self._id

    @property
    def name(self) -> str:
        return self._name

    @name.setter
    @property_not_empty
    def name(self, value: str) -> None:
        self._name = value

    @property
    def owner_id(self) -> Optional[str]:
        return self._owner_id

    @owner_id.setter
    def owner_id(self, value: str) -> None:
        raise NotImplementedError("REST API does not currently support updating project owner.")

    def is_default(self):
        return self.name.lower() == "default"

    def _parse_common_tags(self, project_xml, ns):
        if not isinstance(project_xml, ET.Element):
            project_xml = fromstring(project_xml).find(".//t:project", namespaces=ns)

        if project_xml is not None:
            (
                _,
                name,
                description,
                content_permissions,
                parent_id,
            ) = self._parse_element(project_xml)
            self._set_values(None, name, description, content_permissions, parent_id)
        return self

    def _set_values(self, project_id, name, description, content_permissions, parent_id, owner_id):
        if project_id is not None:
            self._id = project_id
        if name:
            self._name = name
        if description:
            self.description = description
        if content_permissions:
            self._content_permissions = content_permissions
        if parent_id:
            self.parent_id = parent_id
        if owner_id:
            self._owner_id = owner_id

    def _set_permissions(self, permissions):
        self._permissions = permissions

    def _set_default_permissions(self, permissions, content_type):
        attr = "_default_{content}_permissions".format(content=content_type)
        setattr(
            self,
            attr,
            permissions,
        )
        fetch_call = getattr(self, attr)
        logging.getLogger().info({"type": attr, "value": fetch_call()})
        return fetch_call()

    @classmethod
    def from_response(cls, resp, ns) -> List["ProjectItem"]:
        all_project_items = list()
        parsed_response = fromstring(resp)
        all_project_xml = parsed_response.findall(".//t:project", namespaces=ns)

        for project_xml in all_project_xml:
            (
                id,
                name,
                description,
                content_permissions,
                parent_id,
                owner_id,
            ) = cls._parse_element(project_xml)
            project_item = cls(name)
            project_item._set_values(id, name, description, content_permissions, parent_id, owner_id)
            all_project_items.append(project_item)
        return all_project_items

    @staticmethod
    def _parse_element(project_xml):
        id = project_xml.get("id", None)
        name = project_xml.get("name", None)
        description = project_xml.get("description", None)
        content_permissions = project_xml.get("contentPermissions", None)
        parent_id = project_xml.get("parentProjectId", None)
        owner_id = None
        for owner in project_xml:
            owner_id = owner.get("id", None)

        return id, name, description, content_permissions, parent_id, owner_id



import xml.etree.ElementTree as ET
from typing import Set

from defusedxml.ElementTree import fromstring


class TagItem(object):
    @classmethod
    def from_response(cls, resp: bytes, ns) -> Set[str]:
        return cls.from_xml_element(fromstring(resp), ns)

    @classmethod
    def from_xml_element(cls, parsed_response: ET.Element, ns) -> Set[str]:
        all_tags = set()
        tag_elem = parsed_response.findall(".//t:tag", namespaces=ns)
        for tag_xml in tag_elem:
            tag = tag_xml.get("label", None)
            if tag is not None:
                all_tags.add(tag)
        return all_tags



import logging

from defusedxml.ElementTree import fromstring

from .datasource_item import DatasourceItem
from .flow_item import FlowItem
from .project_item import ProjectItem
from .view_item import ViewItem
from .workbook_item import WorkbookItem

logger = logging.getLogger("tableau.models.favorites_item")

from typing import Dict, List, Union

FavoriteType = Dict[
    str,
    List[
        Union[
            DatasourceItem,
            ProjectItem,
            FlowItem,
            ViewItem,
            WorkbookItem,
        ]
    ],
]


class FavoriteItem:
    class Type:
        Workbook: str = "workbook"
        Datasource: str = "datasource"
        View: str = "view"
        Project: str = "project"
        Flow: str = "flow"

    @classmethod
    def from_response(cls, xml: str, namespace: Dict) -> FavoriteType:
        favorites: FavoriteType = {
            "datasources": [],
            "flows": [],
            "projects": [],
            "views": [],
            "workbooks": [],
        }

        parsed_response = fromstring(xml)
        for workbook in parsed_response.findall(".//t:favorite/t:workbook", namespace):
            fav_workbook = WorkbookItem("")
            fav_workbook._set_values(*fav_workbook._parse_element(workbook, namespace))
            if fav_workbook:
                favorites["workbooks"].append(fav_workbook)
        for view in parsed_response.findall(".//t:favorite[t:view]", namespace):
            fav_views = ViewItem.from_xml_element(view, namespace)
            if fav_views:
                for fav_view in fav_views:
                    favorites["views"].append(fav_view)
        for datasource in parsed_response.findall(".//t:favorite/t:datasource", namespace):
            fav_datasource = DatasourceItem("")
            fav_datasource._set_values(*fav_datasource._parse_element(datasource, namespace))
            if fav_datasource:
                favorites["datasources"].append(fav_datasource)
        for project in parsed_response.findall(".//t:favorite/t:project", namespace):
            fav_project = ProjectItem("p")
            fav_project._set_values(*fav_project._parse_element(project))
            if fav_project:
                favorites["projects"].append(fav_project)
        for flow in parsed_response.findall(".//t:favorite/t:flow", namespace):
            fav_flow = FlowItem("flows")
            fav_flow._set_values(*fav_flow._parse_element(flow, namespace))
            if fav_flow:
                favorites["flows"].append(fav_flow)

        return favorites



# coding=utf-8

import unittest

import tableauserverclient as TSC


class SiteModelTests(unittest.TestCase):
    def test_invalid_name(self):
        self.assertRaises(ValueError, TSC.SiteItem, None, "url")
        self.assertRaises(ValueError, TSC.SiteItem, "", "url")
        site = TSC.SiteItem("site", "url")
        with self.assertRaises(ValueError):
            site.name = None

        with self.assertRaises(ValueError):
            site.name = ""

    def test_invalid_admin_mode(self):
        site = TSC.SiteItem("site", "url")
        with self.assertRaises(ValueError):
            site.admin_mode = "Hello"

    def test_invalid_content_url(self):
        with self.assertRaises(ValueError):
            site = TSC.SiteItem(name="蚵仔煎", content_url="蚵仔煎")

        with self.assertRaises(ValueError):
            site = TSC.SiteItem(name="蚵仔煎", content_url=None)

    def test_set_valid_content_url(self):
        # Default Site
        site = TSC.SiteItem(name="Default", content_url="")
        self.assertEqual(site.content_url, "")

        # Unicode Name and ascii content_url
        site = TSC.SiteItem(name="蚵仔煎", content_url="omlette")
        self.assertEqual(site.content_url, "omlette")

    def test_invalid_disable_subscriptions(self):
        site = TSC.SiteItem("site", "url")
        with self.assertRaises(ValueError):
            site.disable_subscriptions = "Hello"

        with self.assertRaises(ValueError):
            site.disable_subscriptions = None

    def test_invalid_revision_history_enabled(self):
        site = TSC.SiteItem("site", "url")
        with self.assertRaises(ValueError):
            site.revision_history_enabled = "Hello"

        with self.assertRaises(ValueError):
            site.revision_history_enabled = None

    def test_invalid_state(self):
        site = TSC.SiteItem("site", "url")
        with self.assertRaises(ValueError):
            site.state = "Hello"

    def test_invalid_subscribe_others_enabled(self):
        site = TSC.SiteItem("site", "url")
        with self.assertRaises(ValueError):
            site.subscribe_others_enabled = "Hello"

        with self.assertRaises(ValueError):
            site.subscribe_others_enabled = None



from defusedxml.ElementTree import fromstring


class FileuploadItem(object):
    def __init__(self):
        self._file_size = None
        self._upload_session_id = None

    @property
    def upload_session_id(self):
        return self._upload_session_id

    @property
    def file_size(self):
        return self._file_size

    @classmethod
    def from_response(cls, resp, ns):
        parsed_response = fromstring(resp)
        fileupload_elem = parsed_response.find(".//t:fileUpload", namespaces=ns)
        fileupload_item = cls()
        fileupload_item._upload_session_id = fileupload_elem.get("uploadSessionId", None)
        fileupload_item._file_size = fileupload_elem.get("fileSize", None)
        return fileupload_item



import logging
from typing import List, Optional

from defusedxml.ElementTree import fromstring

from .connection_credentials import ConnectionCredentials
from .property_decorators import property_is_boolean


class ConnectionItem(object):
    def __init__(self):
        self._datasource_id: Optional[str] = None
        self._datasource_name: Optional[str] = None
        self._id: Optional[str] = None
        self._connection_type: Optional[str] = None
        self.embed_password: bool = None
        self.password: Optional[str] = None
        self.server_address: Optional[str] = None
        self.server_port: Optional[str] = None
        self.username: Optional[str] = None
        self.connection_credentials: Optional[ConnectionCredentials] = None
        self._query_tagging: Optional[bool] = None

    @property
    def datasource_id(self) -> Optional[str]:
        return self._datasource_id

    @property
    def datasource_name(self) -> Optional[str]:
        return self._datasource_name

    @property
    def id(self) -> Optional[str]:
        return self._id

    @property
    def connection_type(self) -> Optional[str]:
        return self._connection_type

    @property
    def query_tagging(self) -> Optional[bool]:
        return self._query_tagging

    @query_tagging.setter
    @property_is_boolean
    def query_tagging(self, value: Optional[bool]):
        # if connection type = hyper, Snowflake, or Teradata, we can't change this value: it is always true
        if self._connection_type in ["hyper", "snowflake", "teradata"]:
            logger = logging.getLogger("tableauserverclient.models.connection_item")
            logger.debug(
                "Cannot update value: Query tagging is always enabled for {} connections".format(self._connection_type)
            )
            return
        self._query_tagging = value

    def __repr__(self):
        return "<ConnectionItem#{_id} embed={embed_password} type={_connection_type} username={username}>".format(
            **self.__dict__
        )

    @classmethod
    def from_response(cls, resp, ns) -> List["ConnectionItem"]:
        all_connection_items = list()
        parsed_response = fromstring(resp)
        all_connection_xml = parsed_response.findall(".//t:connection", namespaces=ns)
        for connection_xml in all_connection_xml:
            connection_item = cls()
            connection_item._id = connection_xml.get("id", None)
            connection_item._connection_type = connection_xml.get("type", None)
            connection_item.embed_password = string_to_bool(connection_xml.get("embedPassword", ""))
            connection_item.server_address = connection_xml.get("serverAddress", None)
            connection_item.server_port = connection_xml.get("serverPort", None)
            connection_item.username = connection_xml.get("userName", None)
            connection_item._query_tagging = string_to_bool(connection_xml.get("queryTaggingEnabled", None))
            datasource_elem = connection_xml.find(".//t:datasource", namespaces=ns)
            if datasource_elem is not None:
                connection_item._datasource_id = datasource_elem.get("id", None)
                connection_item._datasource_name = datasource_elem.get("name", None)
            all_connection_items.append(connection_item)
        return all_connection_items

    @classmethod
    def from_xml_element(cls, parsed_response, ns) -> List["ConnectionItem"]:
        """
        <connections>
            <connection serverAddress="mysql.test.com">
                <connectionCredentials embed="true" name="test" password="secret" />
            </connection>
            <connection serverAddress="pgsql.test.com">
                <connectionCredentials embed="true" name="test" password="secret" />
                </connection>
        </connections>
        """
        all_connection_items: List["ConnectionItem"] = list()
        all_connection_xml = parsed_response.findall(".//t:connection", namespaces=ns)

        for connection_xml in all_connection_xml:
            connection_item = cls()

            connection_item.server_address = connection_xml.get("serverAddress", None)
            connection_item.server_port = connection_xml.get("serverPort", None)

            connection_credentials = connection_xml.find(".//t:connectionCredentials", namespaces=ns)

            if connection_credentials is not None:
                connection_item.connection_credentials = ConnectionCredentials.from_xml_element(
                    connection_credentials, ns
                )

        return all_connection_items


# Used to convert string represented boolean to a boolean type
def string_to_bool(s: str) -> bool:
    return s is not None and s.lower() == "true"



import datetime
from typing import List, Optional

from defusedxml.ElementTree import fromstring

from tableauserverclient.datetime_helpers import parse_datetime
from .flow_run_item import FlowRunItem


class JobItem(object):
    class FinishCode:
        """
        Status codes as documented on
        https://help.tableau.com/current/api/rest_api/en-us/REST/rest_api_ref_jobs_tasks_and_schedules.htm#query_job
        """

        Success: int = 0
        Failed: int = 1
        Cancelled: int = 2

    def __init__(
        self,
        id_: str,
        job_type: str,
        progress: str,
        created_at: datetime.datetime,
        started_at: Optional[datetime.datetime] = None,
        completed_at: Optional[datetime.datetime] = None,
        finish_code: int = 0,
        notes: Optional[List[str]] = None,
        mode: Optional[str] = None,
        workbook_id: Optional[str] = None,
        datasource_id: Optional[str] = None,
        flow_run: Optional[FlowRunItem] = None,
        updated_at: Optional[datetime.datetime] = None,
    ):
        self._id = id_
        self._type = job_type
        self._progress = progress
        self._created_at = created_at
        self._started_at = started_at
        self._completed_at = completed_at
        self._finish_code = finish_code
        self._notes: List[str] = notes or []
        self._mode = mode
        self._workbook_id = workbook_id
        self._datasource_id = datasource_id
        self._flow_run = flow_run
        self._updated_at = updated_at

    @property
    def id(self) -> str:
        return self._id

    @property
    def type(self) -> str:
        return self._type

    @property
    def progress(self) -> str:
        return self._progress

    @property
    def created_at(self) -> datetime.datetime:
        return self._created_at

    @property
    def started_at(self) -> Optional[datetime.datetime]:
        return self._started_at

    @property
    def completed_at(self) -> Optional[datetime.datetime]:
        return self._completed_at

    @property
    def finish_code(self) -> int:
        return self._finish_code

    @property
    def notes(self) -> List[str]:
        return self._notes

    @property
    def mode(self) -> Optional[str]:
        return self._mode

    @mode.setter
    def mode(self, value: str) -> None:
        # check for valid data here
        self._mode = value

    @property
    def workbook_id(self) -> Optional[str]:
        return self._workbook_id

    @workbook_id.setter
    def workbook_id(self, value: Optional[str]) -> None:
        self._workbook_id = value

    @property
    def datasource_id(self) -> Optional[str]:
        return self._datasource_id

    @datasource_id.setter
    def datasource_id(self, value: Optional[str]) -> None:
        self._datasource_id = value

    @property
    def flow_run(self):
        return self._flow_run

    @flow_run.setter
    def flow_run(self, value):
        self._flow_run = value

    @property
    def updated_at(self) -> Optional[datetime.datetime]:
        return self._updated_at

    def __repr__(self):
        return (
            "<Job#{_id} {_type} created_at({_created_at}) started_at({_started_at}) updated_at({_updated_at}) completed_at({_completed_at})"
            " progress ({_progress}) finish_code({_finish_code})>".format(**self.__dict__)
        )

    @classmethod
    def from_response(cls, xml, ns) -> List["JobItem"]:
        parsed_response = fromstring(xml)
        all_tasks_xml = parsed_response.findall(".//t:job", namespaces=ns)

        all_tasks = [JobItem._parse_element(x, ns) for x in all_tasks_xml]

        return all_tasks

    @classmethod
    def _parse_element(cls, element, ns):
        id_ = element.get("id", None)
        type_ = element.get("type", None)
        progress = element.get("progress", None)
        created_at = parse_datetime(element.get("createdAt", None))
        started_at = parse_datetime(element.get("startedAt", None))
        completed_at = parse_datetime(element.get("completedAt", None))
        finish_code = int(element.get("finishCode", -1))
        notes = [note.text for note in element.findall(".//t:notes", namespaces=ns)] or None
        mode = element.get("mode", None)
        workbook = element.find(".//t:workbook[@id]", namespaces=ns)
        workbook_id = workbook.get("id") if workbook is not None else None
        datasource = element.find(".//t:datasource[@id]", namespaces=ns)
        datasource_id = datasource.get("id") if datasource is not None else None
        flow_run = None
        updated_at = parse_datetime(element.get("updatedAt", None))
        for flow_job in element.findall(".//t:runFlowJobType", namespaces=ns):
            flow_run = FlowRunItem()
            flow_run._id = flow_job.get("flowRunId", None)
            for flow in flow_job.findall(".//t:flow", namespaces=ns):
                flow_run._flow_id = flow.get("id", None)
                flow_run._started_at = created_at or started_at
        return cls(
            id_,
            type_,
            progress,
            created_at,
            started_at,
            completed_at,
            finish_code,
            notes,
            mode,
            workbook_id,
            datasource_id,
            flow_run,
            updated_at,
        )


class BackgroundJobItem(object):
    class Status:
        Pending: str = "Pending"
        InProgress: str = "InProgress"
        Success: str = "Success"
        Failed: str = "Failed"
        Cancelled: str = "Cancelled"

    def __init__(
        self,
        id_: str,
        created_at: datetime.datetime,
        priority: int,
        job_type: str,
        status: str,
        title: Optional[str] = None,
        subtitle: Optional[str] = None,
        started_at: Optional[datetime.datetime] = None,
        ended_at: Optional[datetime.datetime] = None,
    ):
        self._id = id_
        self._type = job_type
        self._status = status
        self._created_at = created_at
        self._started_at = started_at
        self._ended_at = ended_at
        self._priority = priority
        self._title = title
        self._subtitle = subtitle

    @property
    def id(self) -> str:
        return self._id

    @property
    def name(self) -> Optional[str]:
        """For API consistency - all other resource endpoints have a name attribute which is used to display what
        they are.  Alias title as name to allow consistent handling of resources in the list sample."""
        return self._title

    @property
    def status(self) -> str:
        return self._status

    @property
    def type(self) -> str:
        return self._type

    @property
    def created_at(self) -> datetime.datetime:
        return self._created_at

    @property
    def started_at(self) -> Optional[datetime.datetime]:
        return self._started_at

    @property
    def ended_at(self) -> Optional[datetime.datetime]:
        return self._ended_at

    @property
    def title(self) -> Optional[str]:
        return self._title

    @property
    def subtitle(self) -> Optional[str]:
        return self._subtitle

    @property
    def priority(self) -> int:
        return self._priority

    @classmethod
    def from_response(cls, xml, ns) -> List["BackgroundJobItem"]:
        parsed_response = fromstring(xml)
        all_tasks_xml = parsed_response.findall(".//t:backgroundJob", namespaces=ns)
        return [cls._parse_element(x, ns) for x in all_tasks_xml]

    @classmethod
    def _parse_element(cls, element, ns):
        id_ = element.get("id", None)
        type_ = element.get("jobType", None)
        status = element.get("status", None)
        created_at = parse_datetime(element.get("createdAt", None))
        started_at = parse_datetime(element.get("startedAt", None))
        ended_at = parse_datetime(element.get("endedAt", None))
        priority = element.get("priority", None)
        title = element.get("title", None)
        subtitle = element.get("subtitle", None)

        return cls(
            id_,
            created_at,
            priority,
            type_,
            status,
            title,
            subtitle,
            started_at,
            ended_at,
        )



import copy
import logging
from typing import List, Optional, Tuple

from .endpoint import QuerysetEndpoint, api
from .exceptions import MissingRequiredFieldError, ServerResponseError
from tableauserverclient.server import RequestFactory, RequestOptions
from tableauserverclient.models import UserItem, WorkbookItem, PaginationItem, GroupItem
from ..pager import Pager

logger = logging.getLogger("tableau.endpoint.users")


class Users(QuerysetEndpoint):
    @property
    def baseurl(self) -> str:
        return "{0}/sites/{1}/users".format(self.parent_srv.baseurl, self.parent_srv.site_id)

    # Gets all users
    @api(version="2.0")
    def get(self, req_options: Optional[RequestOptions] = None) -> Tuple[List[UserItem], PaginationItem]:
        logger.info("Querying all users on site")

        if req_options is None:
            req_options = RequestOptions()
        req_options._all_fields = True

        url = self.baseurl
        server_response = self.get_request(url, req_options)
        pagination_item = PaginationItem.from_response(server_response.content, self.parent_srv.namespace)
        all_user_items = UserItem.from_response(server_response.content, self.parent_srv.namespace)
        return all_user_items, pagination_item

    # Gets 1 user by id
    @api(version="2.0")
    def get_by_id(self, user_id: str) -> UserItem:
        if not user_id:
            error = "User ID undefined."
            raise ValueError(error)
        logger.info("Querying single user (ID: {0})".format(user_id))
        url = "{0}/{1}".format(self.baseurl, user_id)
        server_response = self.get_request(url)
        return UserItem.from_response(server_response.content, self.parent_srv.namespace).pop()

    # Update user
    @api(version="2.0")
    def update(self, user_item: UserItem, password: Optional[str] = None) -> UserItem:
        if not user_item.id:
            error = "User item missing ID."
            raise MissingRequiredFieldError(error)

        url = "{0}/{1}".format(self.baseurl, user_item.id)
        update_req = RequestFactory.User.update_req(user_item, password)
        server_response = self.put_request(url, update_req)
        logger.info("Updated user item (ID: {0})".format(user_item.id))
        updated_item = copy.copy(user_item)
        return updated_item._parse_common_tags(server_response.content, self.parent_srv.namespace)

    # Delete 1 user by id
    @api(version="2.0")
    def remove(self, user_id: str, map_assets_to: Optional[str] = None) -> None:
        if not user_id:
            error = "User ID undefined."
            raise ValueError(error)
        url = "{0}/{1}".format(self.baseurl, user_id)
        if map_assets_to is not None:
            url += f"?mapAssetsTo={map_assets_to}"
        self.delete_request(url)
        logger.info("Removed single user (ID: {0})".format(user_id))

    # Add new user to site
    @api(version="2.0")
    def add(self, user_item: UserItem) -> UserItem:
        url = self.baseurl
        logger.info("Add user {}".format(user_item.name))
        add_req = RequestFactory.User.add_req(user_item)
        server_response = self.post_request(url, add_req)
        logger.info(server_response)
        new_user = UserItem.from_response(server_response.content, self.parent_srv.namespace).pop()
        logger.info("Added new user (ID: {0})".format(new_user.id))
        return new_user

    # Add new users to site. This does not actually perform a bulk action, it's syntactic sugar
    @api(version="2.0")
    def add_all(self, users: List[UserItem]):
        created = []
        failed = []
        for user in users:
            try:
                result = self.add(user)
                created.append(result)
            except Exception as e:
                failed.append(user)
        return created, failed

    # helping the user by parsing a file they could have used to add users through the UI
    # line format: Username [required], password, display name, license, admin, publish
    @api(version="2.0")
    def create_from_file(self, filepath: str) -> Tuple[List[UserItem], List[Tuple[UserItem, ServerResponseError]]]:
        created = []
        failed = []
        if not filepath.find("csv"):
            raise ValueError("Only csv files are accepted")

        with open(filepath) as csv_file:
            csv_file.seek(0)  # set to start of file in case it has been read earlier
            line: str = csv_file.readline()
            while line and line != "":
                user: UserItem = UserItem.CSVImport.create_user_from_line(line)
                try:
                    print(user)
                    result = self.add(user)
                    created.append(result)
                except ServerResponseError as serverError:
                    print("failed")
                    failed.append((user, serverError))
                line = csv_file.readline()
        return created, failed

    # Get workbooks for user
    @api(version="2.0")
    def populate_workbooks(self, user_item: UserItem, req_options: Optional[RequestOptions] = None) -> None:
        if not user_item.id:
            error = "User item missing ID."
            raise MissingRequiredFieldError(error)

        def wb_pager():
            return Pager(lambda options: self._get_wbs_for_user(user_item, options), req_options)

        user_item._set_workbooks(wb_pager)

    def _get_wbs_for_user(
        self, user_item: UserItem, req_options: Optional[RequestOptions] = None
    ) -> Tuple[List[WorkbookItem], PaginationItem]:
        url = "{0}/{1}/workbooks".format(self.baseurl, user_item.id)
        server_response = self.get_request(url, req_options)
        logger.info("Populated workbooks for user (ID: {0})".format(user_item.id))
        workbook_item = WorkbookItem.from_response(server_response.content, self.parent_srv.namespace)
        pagination_item = PaginationItem.from_response(server_response.content, self.parent_srv.namespace)
        return workbook_item, pagination_item

    def populate_favorites(self, user_item: UserItem) -> None:
        self.parent_srv.favorites.get(user_item)

    # Get groups for user
    @api(version="3.7")
    def populate_groups(self, user_item: UserItem, req_options: Optional[RequestOptions] = None) -> None:
        if not user_item.id:
            error = "User item missing ID."
            raise MissingRequiredFieldError(error)

        def groups_for_user_pager():
            return Pager(
                lambda options: self._get_groups_for_user(user_item, options),
                req_options,
            )

        user_item._set_groups(groups_for_user_pager)

    def _get_groups_for_user(
        self, user_item: UserItem, req_options: Optional[RequestOptions] = None
    ) -> Tuple[List[GroupItem], PaginationItem]:
        url = "{0}/{1}/groups".format(self.baseurl, user_item.id)
        server_response = self.get_request(url, req_options)
        logger.info("Populated groups for user (ID: {0})".format(user_item.id))
        group_item = GroupItem.from_response(server_response.content, self.parent_srv.namespace)
        pagination_item = PaginationItem.from_response(server_response.content, self.parent_srv.namespace)
        return group_item, pagination_item



# This sample verifies that tableau server client is installed
# and you can run it. It also shows the version of the client.

import tableauserverclient as TSC

server = TSC.Server("Fake-Server-Url", use_server_version=False)
print("Client details:")
print(TSC.server.endpoint.Endpoint._make_common_headers("fake-token", "any-content"))



from datetime import datetime

from defusedxml import ElementTree
from defusedxml.ElementTree import fromstring, tostring
from typing import Callable, List, Optional

from .exceptions import UnpopulatedPropertyError
from .user_item import UserItem
from .view_item import ViewItem
from .workbook_item import WorkbookItem
from ..datetime_helpers import parse_datetime


class CustomViewItem(object):
    def __init__(self, id: Optional[str] = None, name: Optional[str] = None) -> None:
        self._content_url: Optional[str] = None  # ?
        self._created_at: Optional["datetime"] = None
        self._id: Optional[str] = id
        self._image: Optional[Callable[[], bytes]] = None
        self._name: Optional[str] = name
        self._shared: Optional[bool] = False
        self._updated_at: Optional["datetime"] = None

        self._owner: Optional[UserItem] = None
        self._view: Optional[ViewItem] = None
        self._workbook: Optional[WorkbookItem] = None

    def __repr__(self: "CustomViewItem"):
        view_info = ""
        if self._view:
            view_info = " view='{}'".format(self._view.name or self._view.id or "unknown")
        wb_info = ""
        if self._workbook:
            wb_info = " workbook='{}'".format(self._workbook.name or self._workbook.id or "unknown")
        owner_info = ""
        if self._owner:
            owner_info = " owner='{}'".format(self._owner.name or self._owner.id or "unknown")
        return "<CustomViewItem id={} name=`{}`{}{}{}>".format(self.id, self.name, view_info, wb_info, owner_info)

    def _set_image(self, image):
        self._image = image

    @property
    def content_url(self) -> Optional[str]:
        return self._content_url

    @property
    def created_at(self) -> Optional["datetime"]:
        return self._created_at

    @property
    def id(self) -> Optional[str]:
        return self._id

    @property
    def image(self) -> bytes:
        if self._image is None:
            error = "View item must be populated with its png image first."
            raise UnpopulatedPropertyError(error)
        return self._image()

    @property
    def name(self) -> Optional[str]:
        return self._name

    @name.setter
    def name(self, value: str):
        self._name = value

    @property
    def shared(self) -> Optional[bool]:
        return self._shared

    @shared.setter
    def shared(self, value: bool):
        self._shared = value

    @property
    def updated_at(self) -> Optional["datetime"]:
        return self._updated_at

    @property
    def owner(self) -> Optional[UserItem]:
        return self._owner

    @owner.setter
    def owner(self, value: UserItem):
        self._owner = value

    @property
    def workbook(self) -> Optional[WorkbookItem]:
        return self._workbook

    @property
    def view(self) -> Optional[ViewItem]:
        return self._view

    @classmethod
    def from_response(cls, resp, ns, workbook_id="") -> Optional["CustomViewItem"]:
        item = cls.list_from_response(resp, ns, workbook_id)
        if not item or len(item) == 0:
            return None
        else:
            return item[0]

    @classmethod
    def list_from_response(cls, resp, ns, workbook_id="") -> List["CustomViewItem"]:
        return cls.from_xml_element(fromstring(resp), ns, workbook_id)

    """
    <customView
    id="37d015c6-bc28-4c88-989c-72c0a171f7aa"
    name="New name 2"
    createdAt="2016-02-03T23:35:09Z"
    updatedAt="2022-09-28T23:56:01Z"
    shared="false">
      <view id="8e33ff19-a7a4-4aa5-9dd8-a171e2b9c29f" name="circle"/>
      <workbook id="2fbe87c9-a7d8-45bf-b2b3-877a26ec9af5" name="marks and viz types 2"/>
      <owner id="cdfe8548-84c8-418e-9b33-2c0728b2398a" name="workgroupuser"/>
    </customView>
    """

    @classmethod
    def from_xml_element(cls, parsed_response, ns, workbook_id="") -> List["CustomViewItem"]:
        all_view_items = list()
        all_view_xml = parsed_response.findall(".//t:customView", namespaces=ns)
        for custom_view_xml in all_view_xml:
            cv_item = cls()
            view_elem: ElementTree = custom_view_xml.find(".//t:view", namespaces=ns)
            workbook_elem: str = custom_view_xml.find(".//t:workbook", namespaces=ns)
            owner_elem: str = custom_view_xml.find(".//t:owner", namespaces=ns)
            cv_item._created_at = parse_datetime(custom_view_xml.get("createdAt", None))
            cv_item._updated_at = parse_datetime(custom_view_xml.get("updatedAt", None))
            cv_item._content_url = custom_view_xml.get("contentUrl", None)
            cv_item._id = custom_view_xml.get("id", None)
            cv_item._name = custom_view_xml.get("name", None)

            if owner_elem is not None:
                parsed_owners = UserItem.from_response_as_owner(tostring(custom_view_xml), ns)
                if parsed_owners and len(parsed_owners) > 0:
                    cv_item._owner = parsed_owners[0]

            if view_elem is not None:
                parsed_views = ViewItem.from_response(tostring(custom_view_xml), ns)
                if parsed_views and len(parsed_views) > 0:
                    cv_item._view = parsed_views[0]

            if workbook_id:
                cv_item._workbook = WorkbookItem(workbook_id)
            elif workbook_elem is not None:
                parsed_workbooks = WorkbookItem.from_response(tostring(custom_view_xml), ns)
                if parsed_workbooks and len(parsed_workbooks) > 0:
                    cv_item._workbook = parsed_workbooks[0]

            all_view_items.append(cv_item)
        return all_view_items



import logging

from .endpoint import Endpoint, api
from tableauserverclient.server import RequestFactory
from tableauserverclient.models import FileuploadItem

# For when a datasource is over 64MB, break it into 5MB(standard chunk size) chunks
CHUNK_SIZE = 1024 * 1024 * 5  # 5MB

logger = logging.getLogger("tableau.endpoint.fileuploads")


class Fileuploads(Endpoint):
    def __init__(self, parent_srv):
        super(Fileuploads, self).__init__(parent_srv)

    @property
    def baseurl(self):
        return "{0}/sites/{1}/fileUploads".format(self.parent_srv.baseurl, self.parent_srv.site_id)

    @api(version="2.0")
    def initiate(self):
        url = self.baseurl
        server_response = self.post_request(url, "")
        fileupload_item = FileuploadItem.from_response(server_response.content, self.parent_srv.namespace)
        upload_id = fileupload_item.upload_session_id
        logger.info("Initiated file upload session (ID: {0})".format(upload_id))
        return upload_id

    @api(version="2.0")
    def append(self, upload_id, data, content_type):
        url = "{0}/{1}".format(self.baseurl, upload_id)
        server_response = self.put_request(url, data, content_type)
        logger.info("Uploading a chunk to session (ID: {0})".format(upload_id))
        return FileuploadItem.from_response(server_response.content, self.parent_srv.namespace)

    def _read_chunks(self, file):
        file_opened = False
        try:
            file_content = open(file, "rb")
            file_opened = True
        except TypeError:
            file_content = file

        try:
            while True:
                chunked_content = file_content.read(CHUNK_SIZE)
                if not chunked_content:
                    break
                yield chunked_content
        finally:
            if file_opened:
                file_content.close()

    def upload(self, file):
        upload_id = self.initiate()
        for chunk in self._read_chunks(file):
            request, content_type = RequestFactory.Fileupload.chunk_req(chunk)
            fileupload_item = self.append(upload_id, request, content_type)
            logger.info("\tPublished {0}MB".format(fileupload_item.file_size))
        logger.info("File upload finished (ID: {0})".format(upload_id))
        return upload_id



import logging

from tableauserverclient.server import RequestFactory
from tableauserverclient.models import TableauItem, PermissionsRule

from .endpoint import Endpoint
from .exceptions import MissingRequiredFieldError

from typing import Callable, TYPE_CHECKING, List, Optional, Union

logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    from ..server import Server
    from ..request_options import RequestOptions


class _PermissionsEndpoint(Endpoint):
    """Adds permission model to another endpoint

    Tableau permissions model is identical between objects, but they are nested under
    the parent object endpoint (i.e. permissions for workbooks are under
    /workbooks/:id/permission).  This class is meant to be instantiated inside a
    parent endpoint which has these supported endpoints
    """

    def __init__(self, parent_srv: "Server", owner_baseurl: Callable[[], str]) -> None:
        super(_PermissionsEndpoint, self).__init__(parent_srv)

        # owner_baseurl is the baseurl of the parent.  The MUST be a lambda
        # since we don't know the full site URL until we sign in.  If
        # populated without, we will get a sign-in error
        self.owner_baseurl = owner_baseurl

    def __str__(self):
        return "<PermissionsEndpoint baseurl={}>".format(self.owner_baseurl)

    def update(self, resource: TableauItem, permissions: List[PermissionsRule]) -> List[PermissionsRule]:
        url = "{0}/{1}/permissions".format(self.owner_baseurl(), resource.id)
        update_req = RequestFactory.Permission.add_req(permissions)
        response = self.put_request(url, update_req)
        permissions = PermissionsRule.from_response(response.content, self.parent_srv.namespace)
        logger.info("Updated permissions for resource {0}: {1}".format(resource.id, permissions))

        return permissions

    def delete(self, resource: TableauItem, rules: Union[PermissionsRule, List[PermissionsRule]]):
        # Delete is the only endpoint that doesn't take a list of rules
        # so let's fake it to keep it consistent
        # TODO that means we need error handling around the call
        if isinstance(rules, PermissionsRule):
            rules = [rules]

        for rule in rules:
            for capability, mode in rule.capabilities.items():
                "/permissions/groups/group-id/capability-name/capability-mode"
                url = "{0}/{1}/permissions/{2}/{3}/{4}/{5}".format(
                    self.owner_baseurl(),
                    resource.id,
                    rule.grantee.tag_name + "s",
                    rule.grantee.id,
                    capability,
                    mode,
                )

                logger.debug("Removing {0} permission for capability {1}".format(mode, capability))

                self.delete_request(url)

            logger.info(
                "Deleted permission for {0} {1} item {2}".format(rule.grantee.tag_name, rule.grantee.id, resource.id)
            )

    def populate(self, item: TableauItem):
        if not item.id:
            error = "Server item is missing ID. Item must be retrieved from server first."
            raise MissingRequiredFieldError(error)

        def permission_fetcher():
            return self._get_permissions(item)

        item._set_permissions(permission_fetcher)
        logger.info("Populated permissions for item (ID: {0})".format(item.id))

    def _get_permissions(self, item: TableauItem, req_options: Optional["RequestOptions"] = None):
        url = "{0}/{1}/permissions".format(self.owner_baseurl(), item.id)
        server_response = self.get_request(url, req_options)
        permissions = PermissionsRule.from_response(server_response.content, self.parent_srv.namespace)
        logger.info("Permissions for resource {0}: {1}".format(item.id, permissions))

        return permissions



####
# This script demonstrates how to use the Tableau Server Client
# to interact with datasources. It explores the different
# functions that the Server API supports on datasources.
#
# With no flags set, this sample will query all datasources,
# pick one datasource and populate its connections, and update
# the datasource. Adding flags will demonstrate the specific feature
# on top of the general operations.
####

import argparse
import logging

import tableauserverclient as TSC


def main():
    parser = argparse.ArgumentParser(description="Explore datasource functions supported by the Server API.")
    # Common options; please keep those in sync across all samples
    parser.add_argument("--server", "-s", required=True, help="server address")
    parser.add_argument("--site", "-S", help="site name")
    parser.add_argument(
        "--token-name", "-p", required=True, help="name of the personal access token used to sign into the server"
    )
    parser.add_argument(
        "--token-value", "-v", required=True, help="value of the personal access token used to sign into the server"
    )
    parser.add_argument(
        "--logging-level",
        "-l",
        choices=["debug", "info", "error"],
        default="error",
        help="desired logging level (set to error by default)",
    )
    # Options specific to this sample
    parser.add_argument("--publish", metavar="FILEPATH", help="path to datasource to publish")
    parser.add_argument("--download", metavar="FILEPATH", help="path to save downloaded datasource")

    args = parser.parse_args()

    # Set logging level based on user input, or error by default
    logging_level = getattr(logging, args.logging_level.upper())
    logging.basicConfig(level=logging_level)

    # SIGN IN
    tableau_auth = TSC.PersonalAccessTokenAuth(args.token_name, args.token_value, site_id=args.site)
    server = TSC.Server(args.server, use_server_version=True)
    with server.auth.sign_in(tableau_auth):
        # Query projects for use when demonstrating publishing and updating
        all_projects, pagination_item = server.projects.get()
        default_project = next((project for project in all_projects if project.is_default()), None)

        # Publish datasource if publish flag is set (-publish, -p)
        if args.publish:
            if default_project is not None:
                new_datasource = TSC.DatasourceItem(default_project.id)
                new_datasource = server.datasources.publish(
                    new_datasource, args.publish, TSC.Server.PublishMode.Overwrite
                )
                print("Datasource published. ID: {}".format(new_datasource.id))
            else:
                print("Publish failed. Could not find the default project.")

        # Gets all datasource items
        all_datasources, pagination_item = server.datasources.get()
        print("\nThere are {} datasources on site: ".format(pagination_item.total_available))
        print([datasource.name for datasource in all_datasources])

        if all_datasources:
            # Pick one datasource from the list
            sample_datasource = all_datasources[0]

            # Populate connections
            server.datasources.populate_connections(sample_datasource)
            print("\nConnections for {}: ".format(sample_datasource.name))
            print(
                [
                    "{0}({1})".format(connection.id, connection.datasource_name)
                    for connection in sample_datasource.connections
                ]
            )

            # Add some tags to the datasource
            original_tag_set = set(sample_datasource.tags)
            sample_datasource.tags.update("a", "b", "c", "d")
            server.datasources.update(sample_datasource)
            print("\nOld tag set: {}".format(original_tag_set))
            print("New tag set: {}".format(sample_datasource.tags))

            # Delete all tags that were added by setting tags to original
            sample_datasource.tags = original_tag_set
            server.datasources.update(sample_datasource)


if __name__ == "__main__":
    main()



####
# This script demonstrates how to use the metadata API to query information on a published data source
#
# To run the script, you must have installed Python 3.7 or later.
####

import argparse
import logging
from pprint import pprint

import tableauserverclient as TSC


def main():
    parser = argparse.ArgumentParser(description="Use the metadata API to get information on a published data source.")
    # Common options; please keep those in sync across all samples
    parser.add_argument("--server", "-s", required=True, help="server address")
    parser.add_argument("--site", "-S", help="site name")
    parser.add_argument(
        "--token-name", "-n", required=True, help="name of the personal access token used to sign into the server"
    )
    parser.add_argument(
        "--token-value", "-v", required=True, help="value of the personal access token used to sign into the server"
    )
    parser.add_argument(
        "--logging-level",
        "-l",
        choices=["debug", "info", "error"],
        default="error",
        help="desired logging level (set to error by default)",
    )
    # Options specific to this sample
    parser.add_argument(
        "datasource_name",
        nargs="?",
        help="The name of the published datasource. If not present, we query all data sources.",
    )

    args = parser.parse_args()

    # Set logging level based on user input, or error by default
    logging_level = getattr(logging, args.logging_level.upper())
    logging.basicConfig(level=logging_level)

    # Sign in to server
    tableau_auth = TSC.PersonalAccessTokenAuth(args.token_name, args.token_value, site_id=args.site)
    server = TSC.Server(args.server, use_server_version=True)
    with server.auth.sign_in(tableau_auth):
        # Execute the query
        result = server.metadata.query(
            """
            query useMetadataApiToQueryOrdersDatabases($name: String){
                publishedDatasources (filter: {name: $name}) {
                    luid
                    name
                    description
                    projectName
                    fields {
                        name
                    }
                }
            }""",
            {"name": args.datasource_name},
        )

        # Display warnings/errors (if any)
        if result.get("errors"):
            print("### Errors/Warnings:")
            pprint(result["errors"])

        # Print the results
        if result.get("data"):
            print("### Results:")
            pprint(result["data"]["publishedDatasources"])


if __name__ == "__main__":
    main()



import logging
from contextlib import closing

from .endpoint import QuerysetEndpoint, api
from .exceptions import MissingRequiredFieldError
from .permissions_endpoint import _PermissionsEndpoint
from .resource_tagger import _ResourceTagger
from tableauserverclient.models import ViewItem, PaginationItem

logger = logging.getLogger("tableau.endpoint.views")

from typing import Iterator, List, Optional, Tuple, TYPE_CHECKING

if TYPE_CHECKING:
    from ..request_options import (
        RequestOptions,
        CSVRequestOptions,
        PDFRequestOptions,
        ImageRequestOptions,
        ExcelRequestOptions,
    )


class Views(QuerysetEndpoint):
    def __init__(self, parent_srv):
        super(Views, self).__init__(parent_srv)
        self._resource_tagger = _ResourceTagger(parent_srv)
        self._permissions = _PermissionsEndpoint(parent_srv, lambda: self.baseurl)

    # Used because populate_preview_image functionaliy requires workbook endpoint
    @property
    def siteurl(self) -> str:
        return "{0}/sites/{1}".format(self.parent_srv.baseurl, self.parent_srv.site_id)

    @property
    def baseurl(self) -> str:
        return "{0}/views".format(self.siteurl)

    @api(version="2.2")
    def get(
        self, req_options: Optional["RequestOptions"] = None, usage: bool = False
    ) -> Tuple[List[ViewItem], PaginationItem]:
        logger.info("Querying all views on site")
        url = self.baseurl
        if usage:
            url += "?includeUsageStatistics=true"
        server_response = self.get_request(url, req_options)
        pagination_item = PaginationItem.from_response(server_response.content, self.parent_srv.namespace)
        all_view_items = ViewItem.from_response(server_response.content, self.parent_srv.namespace)
        return all_view_items, pagination_item

    @api(version="3.1")
    def get_by_id(self, view_id: str) -> ViewItem:
        if not view_id:
            error = "View item missing ID."
            raise MissingRequiredFieldError(error)
        logger.info("Querying single view (ID: {0})".format(view_id))
        url = "{0}/{1}".format(self.baseurl, view_id)
        server_response = self.get_request(url)
        return ViewItem.from_response(server_response.content, self.parent_srv.namespace)[0]

    @api(version="2.0")
    def populate_preview_image(self, view_item: ViewItem) -> None:
        if not view_item.id or not view_item.workbook_id:
            error = "View item missing ID or workbook ID."
            raise MissingRequiredFieldError(error)

        def image_fetcher():
            return self._get_preview_for_view(view_item)

        view_item._set_preview_image(image_fetcher)
        logger.info("Populated preview image for view (ID: {0})".format(view_item.id))

    def _get_preview_for_view(self, view_item: ViewItem) -> bytes:
        url = "{0}/workbooks/{1}/views/{2}/previewImage".format(self.siteurl, view_item.workbook_id, view_item.id)
        server_response = self.get_request(url)
        image = server_response.content
        return image

    @api(version="2.5")
    def populate_image(self, view_item: ViewItem, req_options: Optional["ImageRequestOptions"] = None) -> None:
        if not view_item.id:
            error = "View item missing ID."
            raise MissingRequiredFieldError(error)

        def image_fetcher():
            return self._get_view_image(view_item, req_options)

        view_item._set_image(image_fetcher)
        logger.info("Populated image for view (ID: {0})".format(view_item.id))

    def _get_view_image(self, view_item: ViewItem, req_options: Optional["ImageRequestOptions"]) -> bytes:
        url = "{0}/{1}/image".format(self.baseurl, view_item.id)
        server_response = self.get_request(url, req_options)
        image = server_response.content
        return image

    @api(version="2.7")
    def populate_pdf(self, view_item: ViewItem, req_options: Optional["PDFRequestOptions"] = None) -> None:
        if not view_item.id:
            error = "View item missing ID."
            raise MissingRequiredFieldError(error)

        def pdf_fetcher():
            return self._get_view_pdf(view_item, req_options)

        view_item._set_pdf(pdf_fetcher)
        logger.info("Populated pdf for view (ID: {0})".format(view_item.id))

    def _get_view_pdf(self, view_item: ViewItem, req_options: Optional["PDFRequestOptions"]) -> bytes:
        url = "{0}/{1}/pdf".format(self.baseurl, view_item.id)
        server_response = self.get_request(url, req_options)
        pdf = server_response.content
        return pdf

    @api(version="2.7")
    def populate_csv(self, view_item: ViewItem, req_options: Optional["CSVRequestOptions"] = None) -> None:
        if not view_item.id:
            error = "View item missing ID."
            raise MissingRequiredFieldError(error)

        def csv_fetcher():
            return self._get_view_csv(view_item, req_options)

        view_item._set_csv(csv_fetcher)
        logger.info("Populated csv for view (ID: {0})".format(view_item.id))

    def _get_view_csv(self, view_item: ViewItem, req_options: Optional["CSVRequestOptions"]) -> Iterator[bytes]:
        url = "{0}/{1}/data".format(self.baseurl, view_item.id)

        with closing(self.get_request(url, request_object=req_options, parameters={"stream": True})) as server_response:
            yield from server_response.iter_content(1024)

    @api(version="3.8")
    def populate_excel(self, view_item: ViewItem, req_options: Optional["ExcelRequestOptions"] = None) -> None:
        if not view_item.id:
            error = "View item missing ID."
            raise MissingRequiredFieldError(error)

        def excel_fetcher():
            return self._get_view_excel(view_item, req_options)

        view_item._set_excel(excel_fetcher)
        logger.info("Populated excel for view (ID: {0})".format(view_item.id))

    def _get_view_excel(self, view_item: ViewItem, req_options: Optional["ExcelRequestOptions"]) -> Iterator[bytes]:
        url = "{0}/{1}/crosstab/excel".format(self.baseurl, view_item.id)

        with closing(self.get_request(url, request_object=req_options, parameters={"stream": True})) as server_response:
            yield from server_response.iter_content(1024)

    @api(version="3.2")
    def populate_permissions(self, item: ViewItem) -> None:
        self._permissions.populate(item)

    @api(version="3.2")
    def update_permissions(self, resource, rules):
        return self._permissions.update(resource, rules)

    @api(version="3.2")
    def delete_permission(self, item, capability_item):
        return self._permissions.delete(item, capability_item)

    # Update view. Currently only tags can be updated
    def update(self, view_item: ViewItem) -> ViewItem:
        if not view_item.id:
            error = "View item missing ID. View must be retrieved from server first."
            raise MissingRequiredFieldError(error)

        self._resource_tagger.update_tags(self.baseurl, view_item)

        # Returning view item to stay consistent with datasource/view update functions
        return view_item



import warnings
import xml.etree.ElementTree as ET

from defusedxml.ElementTree import fromstring

from .property_decorators import (
    property_is_enum,
    property_is_boolean,
    property_matches,
    property_not_empty,
    property_not_nullable,
    property_is_int,
)

VALID_CONTENT_URL_RE = r"^[a-zA-Z0-9_\-]*$"

from typing import List, Optional, Union, TYPE_CHECKING

if TYPE_CHECKING:
    from tableauserverclient.server import Server


class SiteItem(object):
    _user_quota: Optional[int] = None
    _tier_creator_capacity: Optional[int] = None
    _tier_explorer_capacity: Optional[int] = None
    _tier_viewer_capacity: Optional[int] = None

    def __str__(self):
        return (
            "<"
            + __name__
            + ": "
            + (self.name or "unnamed")
            + ", "
            + (self.id or "unknown-id")
            + ", "
            + (self.state or "unknown-state")
            + ">"
        )

    class AdminMode:
        ContentAndUsers: str = "ContentAndUsers"
        ContentOnly: str = "ContentOnly"

    class State:
        Active: str = "Active"
        Suspended: str = "Suspended"

    def __init__(
        self,
        name: str,
        content_url: str,
        admin_mode: Optional[str] = None,
        user_quota: Optional[int] = None,
        storage_quota: Optional[int] = None,
        disable_subscriptions: bool = False,
        subscribe_others_enabled: bool = True,
        revision_history_enabled: bool = False,
        revision_limit: int = 25,
        data_acceleration_mode: Optional[str] = None,
        flows_enabled: bool = True,
        cataloging_enabled: bool = True,
        editing_flows_enabled: bool = True,
        scheduling_flows_enabled: bool = True,
        allow_subscription_attachments: bool = True,
        guest_access_enabled: bool = False,
        cache_warmup_enabled: bool = True,
        commenting_enabled: bool = True,
        extract_encryption_mode: Optional[str] = None,
        request_access_enabled: bool = False,
        run_now_enabled: bool = True,
        tier_explorer_capacity: Optional[int] = None,
        tier_creator_capacity: Optional[int] = None,
        tier_viewer_capacity: Optional[int] = None,
        data_alerts_enabled: bool = True,
        commenting_mentions_enabled: bool = True,
        catalog_obfuscation_enabled: bool = True,
        flow_auto_save_enabled: bool = True,
        web_extraction_enabled: bool = True,
        metrics_content_type_enabled: bool = True,
        notify_site_admins_on_throttle: bool = False,
        authoring_enabled: bool = True,
        custom_subscription_email_enabled: bool = False,
        custom_subscription_email: Union[str, bool] = False,
        custom_subscription_footer_enabled: bool = False,
        custom_subscription_footer: Union[str, bool] = False,
        ask_data_mode: str = "EnabledByDefault",
        named_sharing_enabled: bool = True,
        mobile_biometrics_enabled: bool = False,
        sheet_image_enabled: bool = True,
        derived_permissions_enabled: bool = False,
        user_visibility_mode: str = "FULL",
        use_default_time_zone: bool = True,
        time_zone=None,
        auto_suspend_refresh_enabled: bool = True,
        auto_suspend_refresh_inactivity_window: int = 30,
    ):
        self._admin_mode = None
        self._id: Optional[str] = None
        self._num_users = None
        self._state = None
        self._status_reason = None
        self._storage: Optional[str] = None
        self.user_quota = user_quota
        self.storage_quota = storage_quota
        self.content_url = content_url
        self.disable_subscriptions = disable_subscriptions
        self.name = name
        self.revision_history_enabled = revision_history_enabled
        self.revision_limit = revision_limit
        self.subscribe_others_enabled = subscribe_others_enabled
        self.admin_mode = admin_mode
        self.data_acceleration_mode = data_acceleration_mode
        self.cataloging_enabled = cataloging_enabled
        self.flows_enabled = flows_enabled
        self.editing_flows_enabled = editing_flows_enabled
        self.scheduling_flows_enabled = scheduling_flows_enabled
        self.allow_subscription_attachments = allow_subscription_attachments
        self.guest_access_enabled = guest_access_enabled
        self.cache_warmup_enabled = cache_warmup_enabled
        self.commenting_enabled = commenting_enabled
        self.extract_encryption_mode = extract_encryption_mode
        self.request_access_enabled = request_access_enabled
        self.run_now_enabled = run_now_enabled
        self.tier_explorer_capacity = tier_explorer_capacity
        self.tier_creator_capacity = tier_creator_capacity
        self.tier_viewer_capacity = tier_viewer_capacity
        self.data_alerts_enabled = data_alerts_enabled
        self.commenting_mentions_enabled = commenting_mentions_enabled
        self.catalog_obfuscation_enabled = catalog_obfuscation_enabled
        self.flow_auto_save_enabled = flow_auto_save_enabled
        self.web_extraction_enabled = web_extraction_enabled
        self.metrics_content_type_enabled = metrics_content_type_enabled
        self.notify_site_admins_on_throttle = notify_site_admins_on_throttle
        self.authoring_enabled = authoring_enabled
        self.custom_subscription_footer_enabled = custom_subscription_footer_enabled
        self.custom_subscription_email_enabled = custom_subscription_email_enabled
        self.custom_subscription_email = custom_subscription_email
        self.custom_subscription_footer = custom_subscription_footer
        self.ask_data_mode = ask_data_mode
        self.named_sharing_enabled = named_sharing_enabled
        self.mobile_biometrics_enabled = mobile_biometrics_enabled
        self.sheet_image_enabled = sheet_image_enabled
        self.derived_permissions_enabled = derived_permissions_enabled
        self.user_visibility_mode = user_visibility_mode
        self.use_default_time_zone = use_default_time_zone
        self.time_zone = time_zone
        self.auto_suspend_refresh_enabled = auto_suspend_refresh_enabled
        self.auto_suspend_refresh_inactivity_window = auto_suspend_refresh_inactivity_window

    @property
    def admin_mode(self) -> Optional[str]:
        return self._admin_mode

    @admin_mode.setter
    @property_is_enum(AdminMode)
    def admin_mode(self, value: Optional[str]) -> None:
        self._admin_mode = value

    @property
    def content_url(self) -> str:
        return self._content_url

    @content_url.setter
    @property_not_nullable
    @property_matches(
        VALID_CONTENT_URL_RE,
        "content_url can contain only letters, numbers, dashes, and underscores",
    )
    def content_url(self, value: str) -> None:
        self._content_url = value

    @property
    def disable_subscriptions(self) -> bool:
        return self._disable_subscriptions

    @disable_subscriptions.setter
    @property_is_boolean
    def disable_subscriptions(self, value: bool):
        self._disable_subscriptions = value

    @property
    def id(self) -> Optional[str]:
        return self._id

    @property
    def name(self) -> str:
        return self._name

    @name.setter
    @property_not_empty
    def name(self, value: str):
        self._name = value

    @property
    def num_users(self):
        return self._num_users

    @property
    def revision_history_enabled(self) -> bool:
        return self._revision_history_enabled

    @revision_history_enabled.setter
    @property_is_boolean
    def revision_history_enabled(self, value: bool):
        self._revision_history_enabled = value

    @property
    def revision_limit(self) -> int:
        return self._revision_limit

    @revision_limit.setter
    @property_is_int((2, 10000), allowed=[-1])
    def revision_limit(self, value: int):
        self._revision_limit = value

    @property
    def state(self) -> Optional[str]:
        return self._state

    @state.setter
    @property_is_enum(State)
    def state(self, value: Optional[str]) -> None:
        self._state = value

    @property
    def status_reason(self) -> Optional[str]:
        return self._status_reason

    @property
    def storage(self) -> Optional[str]:
        return self._storage

    @property
    def user_quota(self) -> Optional[int]:
        if any((self.tier_creator_capacity, self.tier_explorer_capacity, self.tier_viewer_capacity)):
            warnings.warn("Tiered license level is set. Returning None for user_quota")
            return None
        else:
            return self._user_quota

    @user_quota.setter
    def user_quota(self, value: Optional[int]) -> None:
        if value is not None and any(
            (self.tier_creator_capacity, self.tier_explorer_capacity, self.tier_viewer_capacity)
        ):
            raise ValueError(
                "User quota conflicts with setting tiered license levels. "
                "Use replace_license_tiers_with_user_quota to set those to None, "
                "and set user_quota to the desired value."
            )
        self._user_quota = value

    @property
    def subscribe_others_enabled(self) -> bool:
        return self._subscribe_others_enabled

    @subscribe_others_enabled.setter
    @property_is_boolean
    def subscribe_others_enabled(self, value: bool) -> None:
        self._subscribe_others_enabled = value

    @property
    def data_acceleration_mode(self) -> Optional[str]:
        return self._data_acceleration_mode

    @data_acceleration_mode.setter
    def data_acceleration_mode(self, value: Optional[str]):
        self._data_acceleration_mode = value

    @property
    def cataloging_enabled(self) -> bool:
        return self._cataloging_enabled

    @cataloging_enabled.setter
    def cataloging_enabled(self, value: bool):
        self._cataloging_enabled = value

    def is_default(self) -> bool:
        return self.name.lower() == "default"

    @staticmethod
    def use_new_flow_settings(parent_srv: "Server") -> bool:
        return parent_srv is not None and parent_srv.check_at_least_version("3.10")

    @property
    def flows_enabled(self) -> bool:
        return self._flows_enabled

    @flows_enabled.setter
    @property_is_boolean
    def flows_enabled(self, value: bool) -> None:
        # Flows Enabled' is not a supported site setting in API Version [3.17].
        # In Version 3.10+ use the more granular settings 'Editing Flows Enabled' and/or 'Scheduling Flows Enabled'
        self._flows_enabled = value

    @property
    def editing_flows_enabled(self) -> bool:
        return self._editing_flows_enabled

    @editing_flows_enabled.setter
    @property_is_boolean
    def editing_flows_enabled(self, value: bool) -> None:
        self._editing_flows_enabled = value

    @property
    def scheduling_flows_enabled(self) -> bool:
        return self._scheduling_flows_enabled

    @scheduling_flows_enabled.setter
    @property_is_boolean
    def scheduling_flows_enabled(self, value: bool):
        self._scheduling_flows_enabled = value

    @property
    def allow_subscription_attachments(self) -> bool:
        return self._allow_subscription_attachments

    @allow_subscription_attachments.setter
    @property_is_boolean
    def allow_subscription_attachments(self, value: bool):
        self._allow_subscription_attachments = value

    @property
    def guest_access_enabled(self) -> bool:
        return self._guest_access_enabled

    @guest_access_enabled.setter
    @property_is_boolean
    def guest_access_enabled(self, value: bool) -> None:
        self._guest_access_enabled = value

    @property
    def cache_warmup_enabled(self) -> bool:
        return self._cache_warmup_enabled

    @cache_warmup_enabled.setter
    @property_is_boolean
    def cache_warmup_enabled(self, value: bool):
        self._cache_warmup_enabled = value

    @property
    def commenting_enabled(self) -> bool:
        return self._commenting_enabled

    @commenting_enabled.setter
    @property_is_boolean
    def commenting_enabled(self, value: bool):
        self._commenting_enabled = value

    @property
    def extract_encryption_mode(self) -> Optional[str]:
        return self._extract_encryption_mode

    @extract_encryption_mode.setter
    def extract_encryption_mode(self, value: Optional[str]):
        self._extract_encryption_mode = value

    @property
    def request_access_enabled(self) -> bool:
        return self._request_access_enabled

    @request_access_enabled.setter
    @property_is_boolean
    def request_access_enabled(self, value: bool) -> None:
        self._request_access_enabled = value

    @property
    def run_now_enabled(self) -> bool:
        return self._run_now_enabled

    @run_now_enabled.setter
    @property_is_boolean
    def run_now_enabled(self, value: bool):
        self._run_now_enabled = value

    @property
    def tier_explorer_capacity(self) -> Optional[int]:
        return self._tier_explorer_capacity

    @tier_explorer_capacity.setter
    def tier_explorer_capacity(self, value: Optional[int]) -> None:
        self._tier_explorer_capacity = value

    @property
    def tier_creator_capacity(self) -> Optional[int]:
        return self._tier_creator_capacity

    @tier_creator_capacity.setter
    def tier_creator_capacity(self, value: Optional[int]) -> None:
        self._tier_creator_capacity = value

    @property
    def tier_viewer_capacity(self) -> Optional[int]:
        return self._tier_viewer_capacity

    @tier_viewer_capacity.setter
    def tier_viewer_capacity(self, value: Optional[int]):
        self._tier_viewer_capacity = value

    @property
    def data_alerts_enabled(self) -> bool:
        return self._data_alerts_enabled

    @data_alerts_enabled.setter
    @property_is_boolean
    def data_alerts_enabled(self, value: bool) -> None:
        self._data_alerts_enabled = value

    @property
    def commenting_mentions_enabled(self) -> bool:
        return self._commenting_mentions_enabled

    @commenting_mentions_enabled.setter
    @property_is_boolean
    def commenting_mentions_enabled(self, value: bool) -> None:
        self._commenting_mentions_enabled = value

    @property
    def catalog_obfuscation_enabled(self) -> bool:
        return self._catalog_obfuscation_enabled

    @catalog_obfuscation_enabled.setter
    @property_is_boolean
    def catalog_obfuscation_enabled(self, value: bool) -> None:
        self._catalog_obfuscation_enabled = value

    @property
    def flow_auto_save_enabled(self) -> bool:
        return self._flow_auto_save_enabled

    @flow_auto_save_enabled.setter
    @property_is_boolean
    def flow_auto_save_enabled(self, value: bool) -> None:
        self._flow_auto_save_enabled = value

    @property
    def web_extraction_enabled(self) -> bool:
        return self._web_extraction_enabled

    @web_extraction_enabled.setter
    @property_is_boolean
    def web_extraction_enabled(self, value: bool) -> None:
        self._web_extraction_enabled = value

    @property
    def metrics_content_type_enabled(self) -> bool:
        return self._metrics_content_type_enabled

    @metrics_content_type_enabled.setter
    @property_is_boolean
    def metrics_content_type_enabled(self, value: bool) -> None:
        self._metrics_content_type_enabled = value

    @property
    def notify_site_admins_on_throttle(self) -> bool:
        return self._notify_site_admins_on_throttle

    @notify_site_admins_on_throttle.setter
    @property_is_boolean
    def notify_site_admins_on_throttle(self, value: bool) -> None:
        self._notify_site_admins_on_throttle = value

    @property
    def authoring_enabled(self) -> bool:
        return self._authoring_enabled

    @authoring_enabled.setter
    @property_is_boolean
    def authoring_enabled(self, value: bool) -> None:
        self._authoring_enabled = value

    @property
    def custom_subscription_email_enabled(self) -> bool:
        return self._custom_subscription_email_enabled

    @custom_subscription_email_enabled.setter
    @property_is_boolean
    def custom_subscription_email_enabled(self, value: bool) -> None:
        self._custom_subscription_email_enabled = value

    @property
    def custom_subscription_email(self) -> Union[str, bool]:
        return self._custom_subscription_email

    @custom_subscription_email.setter
    def custom_subscription_email(self, value: Union[str, bool]):
        self._custom_subscription_email = value

    @property
    def custom_subscription_footer_enabled(self) -> bool:
        return self._custom_subscription_footer_enabled

    @custom_subscription_footer_enabled.setter
    @property_is_boolean
    def custom_subscription_footer_enabled(self, value: bool) -> None:
        self._custom_subscription_footer_enabled = value

    @property
    def custom_subscription_footer(self) -> Union[str, bool]:
        return self._custom_subscription_footer

    @custom_subscription_footer.setter
    def custom_subscription_footer(self, value: Union[str, bool]) -> None:
        self._custom_subscription_footer = value

    @property
    def ask_data_mode(self) -> str:
        return self._ask_data_mode

    @ask_data_mode.setter
    def ask_data_mode(self, value: str) -> None:
        self._ask_data_mode = value

    @property
    def named_sharing_enabled(self) -> bool:
        return self._named_sharing_enabled

    @named_sharing_enabled.setter
    @property_is_boolean
    def named_sharing_enabled(self, value: bool) -> None:
        self._named_sharing_enabled = value

    @property
    def mobile_biometrics_enabled(self) -> bool:
        return self._mobile_biometrics_enabled

    @mobile_biometrics_enabled.setter
    @property_is_boolean
    def mobile_biometrics_enabled(self, value: bool) -> None:
        self._mobile_biometrics_enabled = value

    @property
    def sheet_image_enabled(self) -> bool:
        return self._sheet_image_enabled

    @sheet_image_enabled.setter
    @property_is_boolean
    def sheet_image_enabled(self, value: bool) -> None:
        self._sheet_image_enabled = value

    @property
    def derived_permissions_enabled(self) -> bool:
        return self._derived_permissions_enabled

    @derived_permissions_enabled.setter
    @property_is_boolean
    def derived_permissions_enabled(self, value: bool) -> None:
        self._derived_permissions_enabled = value

    @property
    def user_visibility_mode(self) -> str:
        return self._user_visibility_mode

    @user_visibility_mode.setter
    def user_visibility_mode(self, value: str):
        self._user_visibility_mode = value

    @property
    def use_default_time_zone(self):
        return self._use_default_time_zone

    @use_default_time_zone.setter
    def use_default_time_zone(self, value):
        self._use_default_time_zone = value

    @property
    def time_zone(self):
        return self._time_zone

    @time_zone.setter
    def time_zone(self, value):
        self._time_zone = value

    @property
    def auto_suspend_refresh_inactivity_window(self):
        return self._auto_suspend_refresh_inactivity_window

    @auto_suspend_refresh_inactivity_window.setter
    def auto_suspend_refresh_inactivity_window(self, value):
        self._auto_suspend_refresh_inactivity_window = value

    @property
    def auto_suspend_refresh_enabled(self) -> bool:
        return self._auto_suspend_refresh_enabled

    @auto_suspend_refresh_enabled.setter
    def auto_suspend_refresh_enabled(self, value: bool):
        self._auto_suspend_refresh_enabled = value

    def replace_license_tiers_with_user_quota(self, value: int) -> None:
        self.tier_creator_capacity = None
        self.tier_explorer_capacity = None
        self.tier_viewer_capacity = None
        self.user_quota = value

    def _parse_common_tags(self, site_xml, ns):
        if not isinstance(site_xml, ET.Element):
            site_xml = fromstring(site_xml).find(".//t:site", namespaces=ns)
        if site_xml is not None:
            (
                _,
                name,
                content_url,
                _,
                admin_mode,
                state,
                subscribe_others_enabled,
                disable_subscriptions,
                revision_history_enabled,
                user_quota,
                storage_quota,
                revision_limit,
                num_users,
                storage,
                data_acceleration_mode,
                flows_enabled,
                cataloging_enabled,
                editing_flows_enabled,
                scheduling_flows_enabled,
                allow_subscription_attachments,
                guest_access_enabled,
                cache_warmup_enabled,
                commenting_enabled,
                extract_encryption_mode,
                request_access_enabled,
                run_now_enabled,
                tier_explorer_capacity,
                tier_creator_capacity,
                tier_viewer_capacity,
                data_alerts_enabled,
                commenting_mentions_enabled,
                catalog_obfuscation_enabled,
                flow_auto_save_enabled,
                web_extraction_enabled,
                metrics_content_type_enabled,
                notify_site_admins_on_throttle,
                authoring_enabled,
                custom_subscription_email_enabled,
                custom_subscription_email,
                custom_subscription_footer_enabled,
                custom_subscription_footer,
                ask_data_mode,
                named_sharing_enabled,
                mobile_biometrics_enabled,
                sheet_image_enabled,
                derived_permissions_enabled,
                user_visibility_mode,
                use_default_time_zone,
                time_zone,
                auto_suspend_refresh_enabled,
                auto_suspend_refresh_inactivity_window,
            ) = self._parse_element(site_xml, ns)

            self._set_values(
                None,
                name,
                content_url,
                None,
                admin_mode,
                state,
                subscribe_others_enabled,
                disable_subscriptions,
                revision_history_enabled,
                user_quota,
                storage_quota,
                revision_limit,
                num_users,
                storage,
                data_acceleration_mode,
                flows_enabled,
                cataloging_enabled,
                editing_flows_enabled,
                scheduling_flows_enabled,
                allow_subscription_attachments,
                guest_access_enabled,
                cache_warmup_enabled,
                commenting_enabled,
                extract_encryption_mode,
                request_access_enabled,
                run_now_enabled,
                tier_explorer_capacity,
                tier_creator_capacity,
                tier_viewer_capacity,
                data_alerts_enabled,
                commenting_mentions_enabled,
                catalog_obfuscation_enabled,
                flow_auto_save_enabled,
                web_extraction_enabled,
                metrics_content_type_enabled,
                notify_site_admins_on_throttle,
                authoring_enabled,
                custom_subscription_email_enabled,
                custom_subscription_email,
                custom_subscription_footer_enabled,
                custom_subscription_footer,
                ask_data_mode,
                named_sharing_enabled,
                mobile_biometrics_enabled,
                sheet_image_enabled,
                derived_permissions_enabled,
                user_visibility_mode,
                use_default_time_zone,
                time_zone,
                auto_suspend_refresh_enabled,
                auto_suspend_refresh_inactivity_window,
            )
        return self

    def _set_values(
        self,
        id,
        name,
        content_url,
        status_reason,
        admin_mode,
        state,
        subscribe_others_enabled,
        disable_subscriptions,
        revision_history_enabled,
        user_quota,
        storage_quota,
        revision_limit,
        num_users,
        storage,
        data_acceleration_mode,
        flows_enabled,
        cataloging_enabled,
        editing_flows_enabled,
        scheduling_flows_enabled,
        allow_subscription_attachments,
        guest_access_enabled,
        cache_warmup_enabled,
        commenting_enabled,
        extract_encryption_mode,
        request_access_enabled,
        run_now_enabled,
        tier_explorer_capacity,
        tier_creator_capacity,
        tier_viewer_capacity,
        data_alerts_enabled,
        commenting_mentions_enabled,
        catalog_obfuscation_enabled,
        flow_auto_save_enabled,
        web_extraction_enabled,
        metrics_content_type_enabled,
        notify_site_admins_on_throttle,
        authoring_enabled,
        custom_subscription_email_enabled,
        custom_subscription_email,
        custom_subscription_footer_enabled,
        custom_subscription_footer,
        ask_data_mode,
        named_sharing_enabled,
        mobile_biometrics_enabled,
        sheet_image_enabled,
        derived_permissions_enabled,
        user_visibility_mode,
        use_default_time_zone,
        time_zone,
        auto_suspend_refresh_enabled,
        auto_suspend_refresh_inactivity_window,
    ):
        if id is not None:
            self._id = id
        if name:
            self._name = name
        if content_url:
            self._content_url = content_url
        if status_reason:
            self._status_reason = status_reason
        if admin_mode:
            self._admin_mode = admin_mode
        if state:
            self._state = state
        if subscribe_others_enabled is not None:
            self._subscribe_others_enabled = subscribe_others_enabled
        if disable_subscriptions is not None:
            self._disable_subscriptions = disable_subscriptions
        if revision_history_enabled is not None:
            self._revision_history_enabled = revision_history_enabled
        if user_quota:
            try:
                self.user_quota = user_quota
            except ValueError:
                warnings.warn("Tiered license level is set. Setting user_quota to None.")
                self.user_quota = None
        if storage_quota:
            self.storage_quota = storage_quota
        if revision_limit:
            self.revision_limit = revision_limit
        if num_users:
            self._num_users = num_users
        if storage:
            self._storage = storage
        if data_acceleration_mode:
            self._data_acceleration_mode = data_acceleration_mode
        if flows_enabled is not None:
            self.flows_enabled = flows_enabled
        if cataloging_enabled is not None:
            self.cataloging_enabled = cataloging_enabled
        if editing_flows_enabled is not None:
            self.editing_flows_enabled = editing_flows_enabled
        if scheduling_flows_enabled is not None:
            self.scheduling_flows_enabled = scheduling_flows_enabled
        if allow_subscription_attachments is not None:
            self.allow_subscription_attachments = allow_subscription_attachments
        if guest_access_enabled is not None:
            self.guest_access_enabled = guest_access_enabled
        if cache_warmup_enabled is not None:
            self.cache_warmup_enabled = cache_warmup_enabled
        if commenting_enabled is not None:
            self.commenting_enabled = commenting_enabled
        if extract_encryption_mode is not None:
            self.extract_encryption_mode = extract_encryption_mode
        if request_access_enabled is not None:
            self.request_access_enabled = request_access_enabled
        if run_now_enabled is not None:
            self.run_now_enabled = run_now_enabled
        if tier_explorer_capacity:
            self.tier_explorer_capacity = tier_explorer_capacity
        if tier_creator_capacity:
            self.tier_creator_capacity = tier_creator_capacity
        if tier_viewer_capacity:
            self.tier_viewer_capacity = tier_viewer_capacity
        if data_alerts_enabled is not None:
            self.data_alerts_enabled = data_alerts_enabled
        if commenting_mentions_enabled is not None:
            self.commenting_mentions_enabled = commenting_mentions_enabled
        if catalog_obfuscation_enabled is not None:
            self.catalog_obfuscation_enabled = catalog_obfuscation_enabled
        if flow_auto_save_enabled is not None:
            self.flow_auto_save_enabled = flow_auto_save_enabled
        if web_extraction_enabled is not None:
            self.web_extraction_enabled = web_extraction_enabled
        if metrics_content_type_enabled is not None:
            self.metrics_content_type_enabled = metrics_content_type_enabled
        if notify_site_admins_on_throttle is not None:
            self.notify_site_admins_on_throttle = notify_site_admins_on_throttle
        if authoring_enabled is not None:
            self.authoring_enabled = authoring_enabled
        if custom_subscription_email_enabled is not None:
            self.custom_subscription_email_enabled = custom_subscription_email_enabled
        if custom_subscription_email is not None:
            self.custom_subscription_email = custom_subscription_email
        if custom_subscription_footer_enabled is not None:
            self.custom_subscription_footer_enabled = custom_subscription_footer_enabled
        if custom_subscription_footer is not None:
            self.custom_subscription_footer = custom_subscription_footer
        if ask_data_mode is not None:
            self.ask_data_mode = ask_data_mode
        if named_sharing_enabled is not None:
            self.named_sharing_enabled = named_sharing_enabled
        if mobile_biometrics_enabled is not None:
            self.mobile_biometrics_enabled = mobile_biometrics_enabled
        if sheet_image_enabled is not None:
            self.sheet_image_enabled = sheet_image_enabled
        if derived_permissions_enabled is not None:
            self.derived_permissions_enabled = derived_permissions_enabled
        if user_visibility_mode is not None:
            self.user_visibility_mode = user_visibility_mode
        if use_default_time_zone is not None:
            self.use_default_time_zone = use_default_time_zone
        if time_zone is not None:
            self.time_zone = time_zone
        if auto_suspend_refresh_enabled is not None:
            self.auto_suspend_refresh_enabled = auto_suspend_refresh_enabled
        if auto_suspend_refresh_inactivity_window is not None:
            self.auto_suspend_refresh_inactivity_window = auto_suspend_refresh_inactivity_window

    @classmethod
    def from_response(cls, resp, ns) -> List["SiteItem"]:
        all_site_items = list()
        parsed_response = fromstring(resp)
        all_site_xml = parsed_response.findall(".//t:site", namespaces=ns)
        for site_xml in all_site_xml:
            (
                id,
                name,
                content_url,
                status_reason,
                admin_mode,
                state,
                subscribe_others_enabled,
                disable_subscriptions,
                revision_history_enabled,
                user_quota,
                storage_quota,
                revision_limit,
                num_users,
                storage,
                data_acceleration_mode,
                flows_enabled,
                cataloging_enabled,
                editing_flows_enabled,
                scheduling_flows_enabled,
                allow_subscription_attachments,
                guest_access_enabled,
                cache_warmup_enabled,
                commenting_enabled,
                extract_encryption_mode,
                request_access_enabled,
                run_now_enabled,
                tier_explorer_capacity,
                tier_creator_capacity,
                tier_viewer_capacity,
                data_alerts_enabled,
                commenting_mentions_enabled,
                catalog_obfuscation_enabled,
                flow_auto_save_enabled,
                web_extraction_enabled,
                metrics_content_type_enabled,
                notify_site_admins_on_throttle,
                authoring_enabled,
                custom_subscription_email_enabled,
                custom_subscription_email,
                custom_subscription_footer_enabled,
                custom_subscription_footer,
                ask_data_mode,
                named_sharing_enabled,
                mobile_biometrics_enabled,
                sheet_image_enabled,
                derived_permissions_enabled,
                user_visibility_mode,
                use_default_time_zone,
                time_zone,
                auto_suspend_refresh_enabled,
                auto_suspend_refresh_inactivity_window,
            ) = cls._parse_element(site_xml, ns)

            site_item = cls(name, content_url)
            site_item._set_values(
                id,
                name,
                content_url,
                status_reason,
                admin_mode,
                state,
                subscribe_others_enabled,
                disable_subscriptions,
                revision_history_enabled,
                user_quota,
                storage_quota,
                revision_limit,
                num_users,
                storage,
                data_acceleration_mode,
                flows_enabled,
                cataloging_enabled,
                editing_flows_enabled,
                scheduling_flows_enabled,
                allow_subscription_attachments,
                guest_access_enabled,
                cache_warmup_enabled,
                commenting_enabled,
                extract_encryption_mode,
                request_access_enabled,
                run_now_enabled,
                tier_explorer_capacity,
                tier_creator_capacity,
                tier_viewer_capacity,
                data_alerts_enabled,
                commenting_mentions_enabled,
                catalog_obfuscation_enabled,
                flow_auto_save_enabled,
                web_extraction_enabled,
                metrics_content_type_enabled,
                notify_site_admins_on_throttle,
                authoring_enabled,
                custom_subscription_email_enabled,
                custom_subscription_email,
                custom_subscription_footer_enabled,
                custom_subscription_footer,
                ask_data_mode,
                named_sharing_enabled,
                mobile_biometrics_enabled,
                sheet_image_enabled,
                derived_permissions_enabled,
                user_visibility_mode,
                use_default_time_zone,
                time_zone,
                auto_suspend_refresh_enabled,
                auto_suspend_refresh_inactivity_window,
            )
            all_site_items.append(site_item)
        return all_site_items

    @staticmethod
    def _parse_element(site_xml, ns):
        id = site_xml.get("id", None)
        name = site_xml.get("name", None)
        content_url = site_xml.get("contentUrl", None)
        status_reason = site_xml.get("statusReason", None)
        admin_mode = site_xml.get("adminMode", None)
        state = site_xml.get("state", None)
        subscribe_others_enabled = string_to_bool(site_xml.get("subscribeOthersEnabled", ""))
        disable_subscriptions = string_to_bool(site_xml.get("disableSubscriptions", ""))
        revision_history_enabled = string_to_bool(site_xml.get("revisionHistoryEnabled", ""))
        editing_flows_enabled = string_to_bool(site_xml.get("editingFlowsEnabled", ""))
        scheduling_flows_enabled = string_to_bool(site_xml.get("schedulingFlowsEnabled", ""))
        allow_subscription_attachments = string_to_bool(site_xml.get("allowSubscriptionAttachments", ""))
        guest_access_enabled = string_to_bool(site_xml.get("guestAccessEnabled", ""))
        cache_warmup_enabled = string_to_bool(site_xml.get("cacheWarmupEnabled", ""))
        commenting_enabled = string_to_bool(site_xml.get("commentingEnabled", ""))
        extract_encryption_mode = site_xml.get("extractEncryptionMode", None)
        request_access_enabled = string_to_bool(site_xml.get("requestAccessEnabled", ""))
        run_now_enabled = string_to_bool(site_xml.get("runNowEnabled", ""))
        tier_explorer_capacity = site_xml.get("tierExplorerCapacity", None)
        if tier_explorer_capacity:
            tier_explorer_capacity = int(tier_explorer_capacity)
        tier_creator_capacity = site_xml.get("tierCreatorCapacity", None)
        if tier_creator_capacity:
            tier_creator_capacity = int(tier_creator_capacity)
        tier_viewer_capacity = site_xml.get("tierViewerCapacity", None)
        if tier_viewer_capacity:
            tier_viewer_capacity = int(tier_viewer_capacity)
        data_alerts_enabled = string_to_bool(site_xml.get("dataAlertsEnabled", ""))
        commenting_mentions_enabled = string_to_bool(site_xml.get("commentingMentionsEnabled", ""))
        catalog_obfuscation_enabled = string_to_bool(site_xml.get("catalogObfuscationEnabled", ""))
        flow_auto_save_enabled = string_to_bool(site_xml.get("flowAutoSaveEnabled", ""))
        web_extraction_enabled = string_to_bool(site_xml.get("webExtractionEnabled", ""))
        metrics_content_type_enabled = string_to_bool(site_xml.get("metricsContentTypeEnabled", ""))
        notify_site_admins_on_throttle = string_to_bool(site_xml.get("notifySiteAdminsOnThrottle", ""))
        authoring_enabled = string_to_bool(site_xml.get("authoringEnabled", ""))
        custom_subscription_email_enabled = string_to_bool(site_xml.get("customSubscriptionEmailEnabled", ""))
        custom_subscription_email = site_xml.get("customSubscriptionEmail", None)
        custom_subscription_footer_enabled = string_to_bool(site_xml.get("customSubscriptionFooterEnabled", ""))
        custom_subscription_footer = site_xml.get("customSubscriptionFooter", None)
        ask_data_mode = site_xml.get("askDataMode", None)
        named_sharing_enabled = string_to_bool(site_xml.get("namedSharingEnabled", ""))
        mobile_biometrics_enabled = string_to_bool(site_xml.get("mobileBiometricsEnabled", ""))
        sheet_image_enabled = string_to_bool(site_xml.get("sheetImageEnabled", ""))
        derived_permissions_enabled = string_to_bool(site_xml.get("derivedPermissionsEnabled", ""))
        user_visibility_mode = site_xml.get("userVisibilityMode", "")
        use_default_time_zone = string_to_bool(site_xml.get("useDefaultTimeZone", ""))
        time_zone = site_xml.get("timeZone", None)
        auto_suspend_refresh_enabled = string_to_bool(site_xml.get("autoSuspendRefreshEnabled", ""))
        auto_suspend_refresh_inactivity_window = site_xml.get("autoSuspendRefreshInactivityWindow", None)
        if auto_suspend_refresh_inactivity_window:
            auto_suspend_refresh_inactivity_window = int(auto_suspend_refresh_inactivity_window)

        user_quota = site_xml.get("userQuota", None)
        if user_quota:
            user_quota = int(user_quota)

        storage_quota = site_xml.get("storageQuota", None)
        if storage_quota:
            storage_quota = int(storage_quota)

        revision_limit = site_xml.get("revisionLimit", None)
        if revision_limit:
            revision_limit = int(revision_limit)

        num_users = None
        storage = None
        usage_elem = site_xml.find(".//t:usage", namespaces=ns)
        if usage_elem is not None:
            num_users = usage_elem.get("numUsers", None)
            storage = usage_elem.get("storage", None)

        data_acceleration_mode = site_xml.get("dataAccelerationMode", "")

        flows_enabled = string_to_bool(site_xml.get("flowsEnabled", ""))
        cataloging_enabled = string_to_bool(site_xml.get("catalogingEnabled", ""))

        return (
            id,
            name,
            content_url,
            status_reason,
            admin_mode,
            state,
            subscribe_others_enabled,
            disable_subscriptions,
            revision_history_enabled,
            user_quota,
            storage_quota,
            revision_limit,
            num_users,
            storage,
            data_acceleration_mode,
            flows_enabled,
            cataloging_enabled,
            editing_flows_enabled,
            scheduling_flows_enabled,
            allow_subscription_attachments,
            guest_access_enabled,
            cache_warmup_enabled,
            commenting_enabled,
            extract_encryption_mode,
            request_access_enabled,
            run_now_enabled,
            tier_explorer_capacity,
            tier_creator_capacity,
            tier_viewer_capacity,
            data_alerts_enabled,
            commenting_mentions_enabled,
            catalog_obfuscation_enabled,
            flow_auto_save_enabled,
            web_extraction_enabled,
            metrics_content_type_enabled,
            notify_site_admins_on_throttle,
            authoring_enabled,
            custom_subscription_email_enabled,
            custom_subscription_email,
            custom_subscription_footer_enabled,
            custom_subscription_footer,
            ask_data_mode,
            named_sharing_enabled,
            mobile_biometrics_enabled,
            sheet_image_enabled,
            derived_permissions_enabled,
            user_visibility_mode,
            use_default_time_zone,
            time_zone,
            auto_suspend_refresh_enabled,
            auto_suspend_refresh_inactivity_window,
        )


# Used to convert string represented boolean to a boolean type
def string_to_bool(s: str) -> bool:
    return s.lower() == "true"
